from future import standard_library
standard_library.install_aliases()
from alert_utils.alert_utils_common.metric_collector.event_writer import *
import queue

class MemoryEventWriter(MetricEventWriter):
    '''
        This class is used to mock a in memory event sink.
        Be careful: all the events are stored in the memory queue
    '''

    def __init__(self, app, config):
        super(MemoryEventWriter, self).__init__(app, config)
        max_queue_size = config.get('max_queue_size', 0)
        self.q = queue.Queue(maxsize=max_queue_size)

    def _flush_msg(self, msg):
        self.q.put(msg)

    def pop_msg(self):
        return self.q.get()

    def get_msg_count(self):
        return self.q.qsize()



from __future__ import absolute_import

import os.path

from .error import RestError


__all__ = [
    'get_base_app_name',
    'remove_http_proxy_env_vars',
]


def get_base_app_name():
    """
    Base App name, which this script belongs to.
    """
    import __main__
    main_name = __main__.__file__
    absolute_path = os.path.normpath(main_name)
    parts = absolute_path.split(os.path.sep)
    parts.reverse()
    for key in ("apps", "slave-apps", "master-apps"):
        try:
            idx = parts.index(key)
            if parts[idx + 1] == "etc":
                return parts[idx - 1]
        except (ValueError, IndexError):
            pass
    raise RestError(
        status=500,
        message='Cannot get app name from file: %s' % main_name
    )


def remove_http_proxy_env_vars():
    for k in ("http_proxy", "https_proxy"):
        if k in os.environ:
            del os.environ[k]
        elif k.upper() in os.environ:
            del os.environ[k.upper()]




from __future__ import absolute_import

# EAI fields
from builtins import object
EAI_ACL = 'eai:acl'
EAI_ATTRIBUTES = 'eai:attributes'
EAI_USER = 'eai:userName'
EAI_APP = 'eai:appName'

EAI_FIELD_PREFIX = 'eai:'
EAI_FIELDS = [EAI_ACL, EAI_ATTRIBUTES, EAI_USER, EAI_APP]

# elements of eai:attributes
EAI_ATTRIBUTES_OPTIONAL = 'optionalFields'
EAI_ATTRIBUTES_REQUIRED = 'requiredFields'
EAI_ATTRIBUTES_WILDCARD = 'wildcardFields'


class RestEAI(object):

    def __init__(self, model, user, app, acl=None):
        self.model = model
        default_acl = {
            'owner': user,
            'app': app,
            'global': 1,
            'can_write': 1,
            'modifiable': 1,
            'removable': 1,
            'sharing': 'global',
            'perms': {'read': ['*'], 'write': ['admin']},
        }
        self.acl = acl or default_acl
        self.user = user
        self.app = app
        self.attributes = self._build_attributes()

    @property
    def content(self):
        return {
            EAI_ACL: self.acl,
            EAI_USER: self.user,
            EAI_APP: self.app,
            EAI_ATTRIBUTES: self.attributes,
        }

    def _build_attributes(self):
        optional_fields = []
        required_fields = []
        for field in self.model.fields:
            if field.required:
                required_fields.append(field.name)
            else:
                optional_fields.append(field.name)
        return {
            EAI_ATTRIBUTES_OPTIONAL: optional_fields,
            EAI_ATTRIBUTES_REQUIRED: required_fields,
            EAI_ATTRIBUTES_WILDCARD: [],
        }



from __future__ import absolute_import
from builtins import object
from os import path as op
import os
from . import arf_consts as ac
from . import alert_actions_exceptions as aae
from shutil import copy
from os import linesep
from mako.template import Template
from mako import exceptions
from munch import Munch
from .alert_actions_template import AlertActionsTemplateMgr
from json import loads as jloads
from .alert_actions_helper import write_file
from .alert_actions_merge import remove_alert_from_conf_file


class AlertActionsConfBase(object):
    def __init__(self, input_setting=None, package_path=None, logger=None,
                 template_dir=None, default_settings_file=None,
                 global_settings=None, **kwargs):
        self._alert_conf_name = "alert_actions.conf"
        self._alert_spec_name = "alert_actions.conf.spec"
        self._eventtypes_conf = "eventtypes.conf"
        self._tags_conf = "tags.conf"
        self._app_conf = "app.conf"
        self._all_settings = input_setting
        self._alert_settings = input_setting[ac.MODULAR_ALERTS]
        self._package_path = package_path
        self._logger = logger
        self._global_settings = global_settings

    def get_local_conf_file_path(self, conf_name=None, create_dir_path=True):
        if not self._package_path:
            return None

        if not conf_name:
            conf_name = self._alert_conf_name

        local_path = op.join(self._package_path, "default")
        if not op.exists(local_path) and create_dir_path:
            os.makedirs(local_path)

        return op.join(local_path, conf_name)

    def get_spec_file_path(self, spec_file=None, create_dir_path=True):
        if not self._package_path:
            return None

        if not spec_file:
            spec_file = self._alert_spec_name

        readme_path = op.join(self._package_path, "README")
        if not op.exists(readme_path) and create_dir_path:
            os.makedirs(readme_path)
        return op.join(readme_path, spec_file)

    def get_icon_dir(self, create_dir_path=True):
        icon_dir = op.join(self._package_path, "appserver", "static")
        if not op.exists(icon_dir) and create_dir_path:
            os.makedirs(icon_dir)
        return icon_dir

    def get_icon_name(self, alert):
        return "alert_" + alert[ac.SHORT_NAME] + ".png"

    def get_icon_path(self, alert, create_dir_path=True):
        return op.join(self.get_icon_dir(create_dir_path=create_dir_path),
                       self.get_icon_name(alert))


class AlertActionsConfGeneration(AlertActionsConfBase):
    DEFAULT_CONF_TEMPLATE = "alert_actions.conf.template"
    DEFAULT_SPEC_TEMPLATE = "alert_actions.conf.spec.template"
    DEFAULT_SETTINGS_FILE = "alert_actions_conf_default_settings.json"
    DEFAULT_ALERT_ICON = "alerticon.png"
    DEFAULT_EVENTTYPES_TEMPLATE = "eventtypes.conf.template"
    DEFAULT_TAGS_TEMPLATE = "tags.conf.template"
    DEFAULT_APP_TEMPLATE = "app.conf.template"

    def __init__(self, input_setting=None, package_path=None, logger=None,
                 template_dir=None, default_settings_file=None, **kwargs):
        if not input_setting or not logger:
            msg = 'status="failed", required_args="input_setting, logger"'
            raise aae.AlertActionsInValidArgs(msg)

        super(AlertActionsConfGeneration, self).__init__(
            input_setting=input_setting,
            package_path=package_path, logger=logger, template_dir=template_dir,
            default_settings_file=default_settings_file,
            **kwargs
        )

        self._html_fields = [ac.PARAMETERS, ]
        self._remove_fields = [ac.SHORT_NAME] + self._html_fields
        self._temp_obj = AlertActionsTemplateMgr(template_dir)
        self._default_settings_file = default_settings_file or \
            op.join(op.dirname(op.abspath(__file__)),
                    AlertActionsConfGeneration.DEFAULT_SETTINGS_FILE)

    def generate_conf(self):
        self._logger.info('status="starting", operation="generate", ' +
                          'object="alert_actions.conf", object_type="file"')
        template = Template(filename=op.join(
            self._temp_obj.get_template_dir(),
            AlertActionsConfGeneration.DEFAULT_CONF_TEMPLATE))
        alert_obj = Munch.fromDict(self._alert_settings)
        try:
            final_string = template.render(mod_alerts=alert_obj)
        except:
            print (exceptions.html_error_template().render())
            raise
        text = linesep.join([s.strip() for s in final_string.splitlines()])
        write_file(self._alert_conf_name,
                   self.get_local_conf_file_path(),
                   text,
                   self._logger)
        self._logger.info('status="success", operation="generate", ' +
                          'object="alert_actions.conf", object_type="file"')

    def generate_eventtypes(self):
        self._logger.info('status="starting", operation="generate", ' +
                          'object="eventtypes.conf", object_type="file"')
        template = Template(filename=op.join(
            self._temp_obj.get_template_dir(),
            AlertActionsConfGeneration.DEFAULT_EVENTTYPES_TEMPLATE))
        alert_obj = Munch.fromDict(self._alert_settings)
        final_string = template.render(mod_alerts=alert_obj)
        text = linesep.join([s.strip() for s in final_string.splitlines()])
        file_path = self.get_local_conf_file_path(conf_name=self._eventtypes_conf)
        write_file(self._eventtypes_conf,
                   file_path,
                   text,
                   self._logger)

        # remove the stanza if not checked
        for alert in self._alert_settings:
            if alert.get("active_response") and alert["active_response"].get("sourcetype"):
                continue
            remove_alert_from_conf_file(alert, file_path, self._logger)
        self._logger.info('status="success", operation="generate", ' +
                          'object="eventtypes.conf", object_type="file"')


    def generate_tags(self):
        self._logger.info('status="starting", operation="generate", ' +
                          'object="tags.conf", object_type="file"')
        template = Template(filename=op.join(
            self._temp_obj.get_template_dir(),
            AlertActionsConfGeneration.DEFAULT_TAGS_TEMPLATE))
        alert_obj = Munch.fromDict(self._alert_settings)
        final_string = template.render(mod_alerts=alert_obj)
        text = linesep.join([s.strip() for s in final_string.splitlines()])
        file_path = self.get_local_conf_file_path(conf_name=self._tags_conf)
        write_file(self._tags_conf,
                   file_path,
                   text,
                   self._logger)

        # remove the stanza if not checked
        for alert in self._alert_settings:
            if alert.get("active_response") and alert["active_response"].get("sourcetype"):
                continue
            remove_alert_from_conf_file(alert, file_path, self._logger)
        self._logger.info('status="success", operation="generate", ' +
                          'object="tags.conf", object_type="file"')

    def generate_spec(self):
        self._logger.info('status="starting", operation="generate", ' +
                          'object="alert_actions.conf.spec", object_type="file"')
        template = Template(filename=op.join(
            self._temp_obj.get_template_dir(),
            AlertActionsConfGeneration.DEFAULT_SPEC_TEMPLATE))
        alert_obj = Munch.fromDict(self._alert_settings)
        final_string = template.render(mod_alerts=alert_obj)
        text = linesep.join([s.strip() for s in final_string.splitlines()])
        write_file(self._alert_spec_name,
                   self.get_spec_file_path(),
                   text,
                   self._logger)
        self._logger.info('status="success", operation="generate", ' +
                          'object="alert_actions.conf.spec", object_type="file"')


    def handle(self):
        self.add_default_settings()
        # self.handler_all_icons()
        self.generate_conf()
        self.generate_spec()
        self.generate_eventtypes()
        self.generate_tags()
        # self.generate_app_conf()


    def add_default_settings(self):
        default_settings = None
        with open(self._default_settings_file, 'r') as df:
            default_settings = jloads(df.read())

        for alert in self._alert_settings:
            if ac.ALERT_PROPS not in list(alert.keys()):
                alert[ac.ALERT_PROPS] = {}
            for k, v in list(default_settings.items()):
                if k in list(alert[ac.ALERT_PROPS].keys()):
                    continue

                alert[ac.ALERT_PROPS][k] = v
                self._logger.info('status="success", operation="Add default setting", alert_name="%s", "%s"="%s"',
                                  alert[ac.SHORT_NAME], k, v)


def generate_alert_actions_conf(input_setting=None, package_path=None,
                                logger=None, **kwargs):
    obj = AlertActionsConfGeneration(input_setting=input_setting,
                                     package_path=package_path,
                                     logger=logger,
                                     **kwargs)
    obj.handle()
    return None



from __future__ import absolute_import
from builtins import object
from . import arf_consts as ac
from os import path as op
from os import remove
from munch import Munch
from mako.template import Template
from mako.lookup import TemplateLookup
from . import alert_actions_exceptions as aae
from .alert_actions_template import AlertActionsTemplateMgr
from .alert_actions_helper import write_file
import re


class AlertActionsPyBase(object):

    def __init__(self, input_setting=None, package_path=None, logger=None,
                 template_py=None, lookup_dir=None, global_settings=None,
                 **kwargs):
        self._all_setting = input_setting
        self._logger = logger
        self._package_path = package_path
        self._global_settings = global_settings
        self._current_alert = None
        self._alert_actions_setting = input_setting[ac.MODULAR_ALERTS]
        self._ta_name = self._all_setting.get(ac.SHORT_NAME)
        self._lib_dir = self.get_python_lib_dir_name(self._ta_name)

    def get_python_lib_dir_name(self, app_name):
        space_replace = re.compile('[^\w]+')
        return space_replace.sub('_', app_name.lower())

    def get_alert_py_name(self, helper=""):
        return self._current_alert[ac.SHORT_NAME] + helper + ".py"

    def get_alert_py_path(self, helper=""):
        if not self._package_path:
            return None
        return op.join(self._package_path, "bin", self.get_alert_py_name())

    def get_alert_helper_py_name(self):
        return "modalert_" + self._current_alert[ac.SHORT_NAME] + "_helper.py"

    def get_alert_helper_py_path(self):
        if not self._package_path:
            return None
        return op.join(self._package_path, "bin", self._lib_dir,
                       self.get_alert_helper_py_name())

    def get_declare_py_name(self):
        return self._lib_dir + "_declare.py"

    def get_decalre_py_path(self):
        if not self._package_path:
            return None
        bin_dir = op.join(self._package_path, "bin")
        return op.join(bin_dir, self.get_declare_py_name())

    def get_template_py_files(self):
        bin_dir = op.join(self._package_path, "bin")
        return [
#            op.join(bin_dir, self._lib_dir + "_declare.py"),
            op.join(bin_dir, self._lib_dir + "_declare.pyc"),
            op.join(bin_dir, self._lib_dir + "_declare.pyo"),
            op.join(bin_dir, self._lib_dir, "alert_actions_base.py"),
            op.join(bin_dir, self._lib_dir, "alert_actions_base.pyc"),
            op.join(bin_dir, self._lib_dir, "alert_actions_base.pyo"),
            op.join(bin_dir, self._lib_dir, "cim_actions.py"),
            op.join(bin_dir, self._lib_dir, "cim_actions.pyc"),
            op.join(bin_dir, self._lib_dir, "cim_actions.pyo"),
            op.join(bin_dir, self._lib_dir, "logging_helper.py"),
            op.join(bin_dir, self._lib_dir, "logging_helper.pyc"),
            op.join(bin_dir, self._lib_dir, "logging_helper.pyo"),
            op.join(bin_dir, self._lib_dir, "setup_util_helper.py"),
            op.join(bin_dir, self._lib_dir, "setup_util_helper.pyc"),
            op.join(bin_dir, self._lib_dir, "setup_util_helper.pyo")
        ]


class AlertActionsPyGenerator(AlertActionsPyBase):
    DEFAULT_TEMPLATE_DECLARE_PY = "python_lib_declare.py.template"
    DEFAULT_TEMPLATE_PY = "alert_action.py.template"
    DEFAULT_TEMPLATE_HELPER_PY = "alert_action_helper.py.template"
    CURRENT_DIR = op.dirname(op.abspath(__file__))
    DEFAULT_LOOKUP_DIR = op.join(CURRENT_DIR, "default_py")

    def __init__(self, input_setting=None, package_path=None, logger=None,
                 template_py=None, template_helper_py=None,
                 template_declare_py=None, lookup_dir=None,
                 global_settings=None,
                 **kwargs):
        if not input_setting or not logger:
            msg = 'required_args="input_setting, logger"'
            raise aae.AlertActionsInValidArgs(msg)
        super(AlertActionsPyGenerator, self).__init__(
            input_setting=input_setting, package_path=package_path,
            logger=logger, template_py=template_py, lookup_dir=lookup_dir,
            global_settings=global_settings, **kwargs)

        self._temp_obj = AlertActionsTemplateMgr()
        self._template = None
        self._template_py = template_py or \
            AlertActionsPyGenerator.DEFAULT_TEMPLATE_PY
        self._template_helper_py = template_helper_py or \
            AlertActionsPyGenerator.DEFAULT_TEMPLATE_HELPER_PY
        self._template_declare_py = template_declare_py or \
            AlertActionsPyGenerator.DEFAULT_TEMPLATE_DECLARE_PY
        self._lookup_dir = lookup_dir or \
            AlertActionsPyGenerator.DEFAULT_LOOKUP_DIR
        self._logger.info("template_py=%s lookup_dir=%s", self._template_py,
                          self._lookup_dir)
        self._output = {}
        self.other_setting = kwargs

    def merge_py_code(self, init, new):
        if not init:
            self._logger.info("No previous code, don't merge new parameters in")
            return new

        start = r'\[sample_code_macro:start\]'
        end = r'\[sample_code_macro:end\]'
        start_str = '[sample_code_macro:start]'
        end_str = '[sample_code_macro:end]'
        pattern = re.compile(start + r'((.|[\r\n])*)' + end,
                             re.MULTILINE)
        matched = pattern.search(init)
        if not matched:
            self._logger.info("No macro anymore, don't merge new parameters in")
            return init

        matched = pattern.search(new)
        if matched:
            new_c = matched.group(1)
            return re.sub(start + r'((.|[\r\n])*)' + end,
                          start_str + new_c + end_str,
                          init)

    def gen_py_file(self, one_alert_setting):
        self._current_alert = one_alert_setting
        self.gen_main_py_file()
        self.gen_helper_py_file()


    def gen_main_py_file(self):
        current_dir = op.dirname(op.abspath(__file__))
        lookup_dir = op.join(current_dir, "default_py")
        tmp_lookup = TemplateLookup(directories=[lookup_dir])

        template_path = self._template_py
        if not op.isabs(self._template_py):
            template_path = op.join(self._temp_obj.get_template_dir(),
                                    self._template_py)
        template = Template(filename=template_path, lookup=tmp_lookup)

        # start to render new py file
        rendered_content = None
        settings = None
        if self._global_settings:
            settings = self._global_settings["settings"]
        rendered_content = template.render(
            input=Munch.fromDict(self._all_setting),
            lib_name=self._lib_dir,
            mod_alert=Munch.fromDict(self._current_alert),
            global_settings=Munch.fromDict(settings),
            helper_name=op.splitext(self.get_alert_helper_py_name())[0]
        )

        self._logger.debug('operation="Writing file", file="%s"',
                           self.get_alert_py_path())
        write_file(self.get_alert_py_name(),
                   self.get_alert_py_path(),
                   rendered_content,
                   self._logger)

    def gen_helper_py_file(self):
        current_dir = op.dirname(op.abspath(__file__))
        lookup_dir = op.join(current_dir, "default_py")
        tmp_lookup = TemplateLookup(directories=[lookup_dir])

        template_path = self._template_helper_py
        if not op.isabs(self._template_helper_py):
            template_path = op.join(self._temp_obj.get_template_dir(),
                                    self._template_helper_py)
        template = Template(filename=template_path, lookup=tmp_lookup)

        name = self._current_alert[ac.SHORT_NAME]
        final_content = None
        rendered_content = None
        init_content = None
        if self._current_alert.get("code"):
            init_content = self._current_alert.get("code")

        # start to render new py file
        settings = {}
        if self._global_settings:
            settings = self._global_settings.get("settings", {})
        rendered_content = template.render(
            input=Munch.fromDict(self._all_setting),
            mod_alert=Munch.fromDict(self._current_alert),
            global_settings=Munch.fromDict(settings)
        )

        final_content = self.merge_py_code(init_content, rendered_content)
        self._logger.debug('operation="Writing file", file="%s"',
                           self.get_alert_py_path())
        write_file(self.get_alert_helper_py_name(),
                   self.get_alert_helper_py_path(),
                   final_content,
                   self._logger)
        if not self._output.get(name):
            self._output[name] = {}
        self._output[name][self.get_alert_py_name(helper="_helper")] = final_content

    def handle(self):
        for alert in self._alert_actions_setting:
            self._logger.info('operation="Generate py file", alert_action="%s"',
                              alert[ac.SHORT_NAME])
            self.gen_py_file(alert)


def generate_alert_actions_py_files(input_setting=None, package_path=None,
                                    logger=None, global_settings=None,
                                    **kwargs):
    py_gen = AlertActionsPyGenerator(input_setting=input_setting,
                                     package_path=package_path,
                                     logger=logger,
                                     global_settings=global_settings,
                                     **kwargs)
    py_gen.handle()
    return py_gen._output




from __future__ import print_function
from __future__ import absolute_import
from builtins import str
import csv
import gzip
import sys

try:
    from splunk.clilib.bundle_paths import make_splunkhome_path
except ImportError:
    from splunk.appserver.mrsparkle.lib.util import make_splunkhome_path
sys.path.append(make_splunkhome_path(["etc", "apps", "Splunk_SA_CIM", "lib"]))

from .cim_actions import ModularAction
from solnlib.log import Logs


class ModularAlertBase(ModularAction):
    def __init__(self, alert_name):
        self._alert_name = alert_name
        # self._logger_name = "modalert_" + alert_name
        self._logger_name = alert_name + "_modalert"
        self._logger = Logs().get_logger(self._logger_name)
        super(ModularAlertBase, self).__init__(
            sys.stdin.read(), self._logger, alert_name)

    def log_error(self, msg):
        self._logger.error(msg)

    def log_info(self, msg):
        self._logger.info(msg)

    def log_debug(self, msg):
        self._logger.debug(msg)

    def get_param(self, param_name):
        return self.configuration.get(param_name)

    def dowork(self, result):
        raise NotImplemented()

    def run(self, argv):
        if len(argv) < 2 or argv[1] != "--execute":
            msg = 'Error: argv="{}", expected="--execute"'.format(argv)
            print(msg, file=sys.stderr)
            sys.exit(1)

        try:
            with gzip.open(self.results_file, 'rb') as rh:
                for num, result in enumerate(csv.DictReader(rh)):
                    result.setdefault('rid', str(num))
                    self.update(result)
                    self.dowork(result)
        except Exception as e:
            self._logger.error(self.message(e, 'failure'))
            self._logger.exception("exception=")
            sys.exit(2)



from __future__ import absolute_import

from builtins import str
from builtins import object
import copy
import json
from multiprocessing.pool import ThreadPool

from solnlib.packages.splunklib.binding import HTTPError

from ..rest_handler.schema import RestSchema
from ..rest_handler.handler import RestHandler


__all__ = [
    "GlobalConfigError",
    "Configuration",
    "Inputs",
    "Configs",
    "Settings",
]


class GlobalConfigError(Exception):
    pass


class Configuration(object):
    """
    Splunk Configuration Handler.
    """

    FILTERS = [u"eai:appName", u"eai:acl", u"eai:userName"]
    ENTITY_NAME = u"name"
    SETTINGS = u"settings"
    NOT_FOUND = u"[404]: Not Found"

    def __init__(self, splunkd_client, schema):
        """

        :param splunkd_client: SplunkRestClient
        :param schema:
        """
        self._client = splunkd_client
        self._schema = schema

    def load(self, *args, **kwargs):
        """
        Load all stored configuration for given schema.

        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError()

    def save_stanza(self, item):
        """
        Save configuration with type_name and configuration

        :param item:
        :return: error while save the configuration
        """
        return self._save_configuration(item[0], item[1])

    def save(self, payload):
        """
        Save configuration. Return error while saving.
        It includes creating and updating. That is, it will try to
        update first, then create if NOT FOUND error occurs.

        :param payload: same format with return of ``load``.
        :return:

        Usage::
        >>> from splunk_add_on_ucc_framework.splunktaucclib.global_config import GlobalConfig
        >>> global_config = GlobalConfig()
        >>> payload = {
        >>>    'settings': [
        >>>        {
        >>>            'name': 'proxy',
        >>>            'proxy_host': '1.2.3.4',
        >>>            'proxy_port': '5678',
        >>>        },
        >>>        {
        >>>            'name': 'logging',
        >>>            'level': 'DEBUG',
        >>>        }
        >>>    ]
        >>> }
        >>> global_config.settings.save(payload)
        """
        # expand the payload to task_list
        task_list = []
        for type_name, configurations in list(payload.items()):
            task_list.extend(
                [(type_name, configuration) for configuration in configurations]
            )
        task_len = len(task_list)
        # return empty error list if task list is empty
        if not task_list:
            return []
        task_len = min(8, task_len)
        pool = ThreadPool(processes=task_len)
        errors = pool.map(self.save_stanza, task_list)
        pool.close()
        pool.join()
        return errors

    @property
    def internal_schema(self):
        """
        Get the schema for inputs, configs and settings

        :return:
        """
        return self._schema.inputs + self._schema.configs + self._schema.settings

    def _save_configuration(self, type_name, configuration):
        schema = self._search_configuration_schema(
            type_name, configuration[self.ENTITY_NAME],
        )
        configuration = copy.copy(configuration)
        self._dump_multiple_select(configuration, schema)

        # update
        try:
            self._update(type_name, copy.copy(configuration))
        except HTTPError as exc:
            if self.NOT_FOUND in str(exc):
                # not exists, go to create
                pass
            else:
                return exc
        except Exception as exc:
            return exc
        else:
            return None

        # create
        try:
            self._create(type_name, configuration)
        except Exception as exc:
            return exc
        else:
            return None

    def _create(self, type_name, configuration):
        self._save_endpoint(
            type_name, configuration,
        )

    def _update(self, type_name, configuration):
        name = configuration[self.ENTITY_NAME]
        del configuration[self.ENTITY_NAME]
        self._save_endpoint(
            type_name, configuration, name=name,
        )

    @classmethod
    def _filter_fields(cls, entity):
        for (k, v) in list(entity.items()):
            if k in cls.FILTERS:
                del entity[k]

    def _load_endpoint(self, name, schema):
        query = {
            "output_mode": "json",
            "count": "0",
            "--cred--": "1",
        }
        response = self._client.get(
            RestHandler.path_segment(self._endpoint_path(name)), **query
        )
        body = response.body.read()
        cont = json.loads(body)

        entities = []
        for entry in cont["entry"]:
            entity = entry["content"]
            entity[self.ENTITY_NAME] = entry["name"]
            self._load_multiple_select(entity, schema)
            entities.append(entity)
        return entities

    def _save_endpoint(self, endpoint, content, name=None):
        endpoint = self._endpoint_path(endpoint)
        self._client.post(RestHandler.path_segment(endpoint, name=name), **content)

    @classmethod
    def _load_multiple_select(cls, entity, schema):
        for field in schema:
            field_type = field.get("type")
            value = entity.get(field["field"])
            if field_type != "multipleSelect" or not value:
                continue
            delimiter = field["options"]["delimiter"]
            entity[field["field"]] = value.split(delimiter)

    @classmethod
    def _dump_multiple_select(cls, entity, schema):
        for field in schema:
            field_type = field.get("type")
            value = entity.get(field["field"])
            if field_type != "multipleSelect" or not value:
                continue
            if not isinstance(value, list):
                continue
            delimiter = field["options"]["delimiter"]
            entity[field["field"]] = delimiter.join(value)

    def _endpoint_path(self, name):
        return "{admin_match}/{endpoint_name}".format(
            admin_match=self._schema.admin_match,
            endpoint_name=RestSchema.endpoint_name(name, self._schema.namespace),
        )

    def _search_configuration_schema(self, type_name, configuration_name):
        for item in self.internal_schema:
            # add support for settings schema
            if item["name"] == type_name or (
                type_name == self.SETTINGS and item["name"] == configuration_name
            ):
                return item["entity"]
        else:
            raise GlobalConfigError(
                "Schema Not Found for Configuration, "
                "configuration_type={configuration_type}, "
                "configuration_name={configuration_name}".format(
                    configuration_type=type_name, configuration_name=configuration_name,
                ),
            )


class Inputs(Configuration):
    def __init__(self, splunkd_client, schema):
        super(Inputs, self).__init__(splunkd_client, schema)
        self._splunkd_client = splunkd_client
        self._schema = schema
        self._references = None

    def load(self, input_type=None):
        """

        :param input_type:
        :return:

        Usage::
        >>> from splunk_add_on_ucc_framework.splunktaucclib.global_config import GlobalConfig
        >>> global_config = GlobalConfig()
        >>> inputs = global_config.inputs.load()
        """
        # move configs read operation out of init method
        if not self._references:
            self._references = Configs(self._splunkd_client, self._schema).load()
        inputs = {}
        for input_item in self.internal_schema:
            if input_type is None or input_item["name"] == input_type:
                input_entities = self._load_endpoint(
                    input_item["name"], input_item["entity"]
                )
                # filter unused fields in response
                for input_entity in input_entities:
                    self._filter_fields(input_entity)
                # expand referenced entity
                self._reference(
                    input_entities, input_item, self._references,
                )
                inputs[input_item["name"]] = input_entities
        return inputs

    @property
    def internal_schema(self):
        return self._schema.inputs

    @classmethod
    def _reference(cls, input_entities, input_item, configs):
        for input_entity in input_entities:
            cls._input_reference(
                input_item["name"], input_entity, input_item["entity"], configs
            )

    @classmethod
    def _input_reference(cls, input_type, input_entity, input_schema, configs):
        for field in input_schema:
            options = field.get("options", {})
            config_type = options.get("referenceName")
            config_name = input_entity.get(field["field"])
            if not config_type or not config_name:
                continue

            for config in configs.get(config_type, []):
                if config["name"] == config_name:
                    input_entity[field["field"]] = config
                    break
            else:
                raise GlobalConfigError(
                    "Config Not Found for Input, "
                    "input_type={input_type}, "
                    "input_name={input_name}, "
                    "config_type={config_type}, "
                    "config_name={config_name}".format(
                        input_type=input_type,
                        input_name=input_entity["name"],
                        config_type=config_type,
                        config_name=config_name,
                    )
                )


class Configs(Configuration):
    def load(self, config_type=None):
        """

        :param config_type:
        :return:

         Usage::
        >>> from splunk_add_on_ucc_framework.splunktaucclib.global_config import GlobalConfig
        >>> global_config = GlobalConfig()
        >>> configs = global_config.configs.load()
        """
        configs = {}
        for config in self.internal_schema:
            if config_type is None or config["name"] == config_type:
                config_entities = self._load_endpoint(config["name"], config["entity"])
                for config_entity in config_entities:
                    self._filter_fields(config_entity)
                configs[config["name"]] = config_entities
        return configs

    @property
    def internal_schema(self):
        return self._schema.configs


class Settings(Configuration):

    TYPE_NAME = u"settings"

    def load(self):
        """

        :return:

         Usage::
        >>> from splunk_add_on_ucc_framework.splunktaucclib.global_config import GlobalConfig
        >>> global_config = GlobalConfig()
        >>> settings = global_config.settings.load()
        """
        settings = []
        for setting in self.internal_schema:
            setting_entity = self._load_endpoint(
                "settings/%s" % setting["name"], setting["entity"]
            )
            self._load_multiple_select(setting_entity[0], setting["entity"])
            entity = setting_entity[0]
            self._filter_fields(entity)
            settings.append(entity)
        return {Settings.TYPE_NAME: settings}

    @property
    def internal_schema(self):
        return self._schema.settings

    def _search_configuration_schema(self, type_name, configuration_name):
        return super(Settings, self)._search_configuration_schema(
            configuration_name, configuration_name,
        )




# encoding = utf-8
from solnlib.splunkenv import make_splunkhome_path

ADDON_BUILDER_APP_NAME = "splunk_app_addon-builder"
ADDON_BUILDER_APP_CATEGORY = "Splunk App Add-on Builder"

# can not use common_util.make_splunk_path here. cylic import
BUILDER_WORKSPACE_ROOT = make_splunkhome_path(
    ['etc', 'apps', ADDON_BUILDER_APP_NAME, 'local', 'builder_workspace'])

# cookie related keys
TA_NAME = 'ta_builder_current_ta_name'
TA_DISPLAY_NAME = 'ta_builder_current_ta_display_name'
TA_WIZARD_STEP = "ta_builder_wizard"
BUILT_FLAG = 'built_by_tabuilder'
COOKIE_KEYS = [TA_NAME, TA_WIZARD_STEP, BUILT_FLAG, TA_DISPLAY_NAME]
COOKIE_EXPIRES_DAY = 30

# data inputs
FIELD_EXTRACTION_MI = "field_extraction_mi"
FIELD_EXTRACTION_MONITOR_MI = "_tab_internal_monitor_inputs_"
VALIDATION_MI = "validation_mi"
VALIDATION_MONITOR_MI = "_tab_internal_monitor_inputs_"

# app inspect
GLOBAL_SETTING_CONF_NAME = "settings"
APP_CERT_STANZA = "app_cert"

# global settings
USR_CREDENTIAL_SETTING = 'credential_settings'
PROXY_SETTING = 'proxy_settings'
LOG_SETTINGS = 'log_settings'
CUSTOMIZED_SETTINGS = 'customized_settings'
CREDENTIAL_SCHEMA = 'default_account'
LOG_SCHEMA = 'default_logging'
PROXY_SCHEMA = 'default_proxy'
CUSTOMIZED_BOOL_SCHEMA = 'bool'
CUSTOMIZED_TEXT_SCHEMA = 'text'
CUSTOMIZED_PASSWORD_SCHEMA = 'password'

CUSTOMIZED_TYPE_MAP = {
    "text": "text",
    "checkbox": "bool",
    "password": "password"
}

# cim mapping settings
CONF_MODELS = "aob_models"
CONF_EVAL_FUNC = "aob_eval_functions"
CONF_CIM_RELATED = ("props.conf", "transforms.conf", "eventtypes.conf", "tags.conf")
CONF_FE_RELATED = ("props.conf", "transforms.conf")

TAB_KO_PREFIX = "aob_gen"

RESERVED_SOURCETYPES = ("splunk:tabuilder:validation",)



"""
Global config schema.
"""

from __future__ import absolute_import

from builtins import map
from builtins import object
import os
import os.path as op
import stat
import shutil
from solnlib.utils import is_true
from splunk_add_on_ucc_framework.splunktaucclib.global_config import GlobalConfigSchema
from splunk_add_on_ucc_framework.splunktaucclib.rest_handler.endpoint.field import (
    RestField,
)
import json

from .endpoint.field import RestFieldBuilder
from .endpoint.single_model import (
    SingleModelEntityBuilder,
    SingleModelEndpointBuilder,
)
from .endpoint.multiple_model import (
    MultipleModelEntityBuilder,
    MultipleModelEndpointBuilder,
)
from .endpoint.datainput import (
    DataInputEndpointBuilder,
    DataInputEntityBuilder,
)

# model to get accesstoken for oauth
from .endpoint.oauth_model import OAuthModelEndpointBuilder
from .endpoint.base import indent, quote_regex


class GlobalConfigBuilderSchema(GlobalConfigSchema):
    def __init__(self, content, j2_env, *args, **kwargs):
        super(GlobalConfigBuilderSchema, self).__init__(content, *args, **kwargs)
        self.j2_env = j2_env
        self._endpoints = {}
        self._parse_builder_schema()

    @property
    def endpoints(self):
        return [endpoint for _, endpoint in list(self._endpoints.items())]

    def _parse_builder_schema(self):
        self._builder_configs()
        self._builder_settings()
        self._builder_inputs()

    def _builder_configs(self):
        # SingleModel
        for config in self._configs:
            self._builder_entity(
                None,
                config["entity"],
                config["name"],
                SingleModelEndpointBuilder,
                SingleModelEntityBuilder,
                conf_name=config.get("conf"),
                rest_handler_name=config.get("restHandlerName"),
            )
            # If we have have given oauth support then we have to add endpoint for accesstoken
            for entity_element in config["entity"]:
                if entity_element["type"] == "oauth":
                    self._get_endpoint(
                        "oauth", OAuthModelEndpointBuilder, app_name=self._meta["name"]
                    )

    def _builder_settings(self):
        # MultipleModel
        for setting in self._settings:
            self._builder_entity(
                setting["name"],
                setting["entity"],
                "settings",
                MultipleModelEndpointBuilder,
                MultipleModelEntityBuilder,
            )

    def _builder_inputs(self):
        # DataInput
        for input_item in self._inputs:
            rest_handler_name = None
            if "restHandlerName" in input_item:
                rest_handler_name = input_item["restHandlerName"]
            if "conf" in input_item:
                self._builder_entity(
                    None,
                    input_item["entity"],
                    input_item["name"],
                    SingleModelEndpointBuilder,
                    SingleModelEntityBuilder,
                    conf_name=input_item["conf"],
                    rest_handler_name=rest_handler_name,
                )
            else:
                self._builder_entity(
                    None,
                    input_item["entity"],
                    input_item["name"],
                    DataInputEndpointBuilder,
                    DataInputEntityBuilder,
                    input_type=input_item["name"],
                    rest_handler_name=rest_handler_name,
                )

    def _builder_entity(
        self, name, content, endpoint, endpoint_builder, entity_builder, *args, **kwargs
    ):
        endpoint_obj = self._get_endpoint(endpoint, endpoint_builder, *args, **kwargs)
        # If the entity contains type oauth then we need to alter the content to generate proper entities to generate
        # the rest handler with the oauth fields
        content = self._get_oauth_enitities(content)
        fields = self._parse_fields(content)
        entity = entity_builder(name, fields, *args, **kwargs)
        endpoint_obj.add_entity(entity)

    def _parse_fields(self, fields_content):
        return [
            self._parse_field(field)
            for field in fields_content
            if field["field"] != "name"
        ]

    def _get_endpoint(self, name, endpoint_builder, *args, **kwargs):
        if name not in self._endpoints:
            endpoint = endpoint_builder(
                name=name,
                namespace=self._meta["restRoot"],
                j2_env=self.j2_env,
                *args,
                **kwargs
            )
            self._endpoints[name] = endpoint
        return self._endpoints[name]

    def _parse_field(self, content):
        field = RestField(
            content["field"],
            required=is_true(content.get("required")),
            encrypted=is_true(content.get("encrypted")),
            default=content.get("defaultValue"),
        )
        return RestFieldBuilder(
            field, self._parse_validation(content.get("validators")),
        )

    def _parse_validation(self, validation):
        global_config_validation = GlobalConfigValidation(validation)
        return global_config_validation.build()

    """
    If the entity contains type oauth then we need to alter the content to generate proper entities to generate
    the rest handler with the oauth fields
    :param content: json content of entity
    :type content: `json`
    """

    def _get_oauth_enitities(self, content):
        for entity_element in content:
            # Check if we have oauth type
            if entity_element["type"] == "oauth":
                # Check if we have both basic and oauth type authentication is required
                if (
                    "basic" in entity_element["options"]["auth_type"]
                    and "oauth" in entity_element["options"]["auth_type"]
                ):
                    # Append all the basic auth fields to the content
                    content = content + entity_element["options"]["basic"]
                    # Append oauth auth fields to the content
                    content = content + entity_element["options"]["oauth"]
                    # Append auth_type, access_token, refresh_token & instance_url fields
                    content = content + json.loads(
                        '[{"field": "access_token","encrypted": true},{"field": "refresh_token","encrypted":true},{"field": "instance_url"},{"field": "auth_type"}]'
                    )
                # If only oauth type authentication is required
                elif "oauth" in entity_element["options"]["auth_type"]:
                    # Append all the oauth auth fields to the content
                    content = content + entity_element["options"]["oauth"]
                    # Append access_token, refresh_token & instance_url fields
                    content = content + json.loads(
                        '[{"field": "access_token","encrypted": true},{"field": "refresh_token","encrypted":true},{"field": "instance_url"}]'
                    )
                # We will remove the oauth type entity as we have replaced it with all the entity fields
                content.remove(entity_element)
                break
        return content


class GlobalConfigValidation(object):

    _validation_template = """validator.{validator}({arguments})"""

    def __init__(self, validation):
        self._validators = []
        self._validation = validation

    def build(self):
        if not self._validation:
            return None
        for item in self._validation:
            parser = GlobalConfigValidation.validation_mapping.get(item["type"], None)
            if parser is None:
                continue
            validator, arguments = parser(item)
            if validator is None:
                continue
            arguments = arguments or {}
            self._validators.append(
                self._validation_template.format(
                    validator=validator, arguments=self._arguments(**arguments),
                )
            )

        if not self._validators:
            return None
        if len(self._validators) > 1:
            return self.multiple_validators(self._validators)
        else:
            return self._validators[0]

    @classmethod
    def _arguments(cls, **kwargs):
        if not kwargs:
            return ""
        args = list(
            map(lambda k_v: "{}={}, ".format(k_v[0], k_v[1]), list(kwargs.items()),)
        )
        args.insert(0, "")
        args.append("")
        return indent("\n".join(args))

    @classmethod
    def _content(cls, validator, arguments):
        pass

    @classmethod
    def string(cls, validation):
        return (
            "String",
            {
                "max_len": validation.get("maxLength"),
                "min_len": validation.get("minLength"),
            },
        )

    @classmethod
    def number(cls, validation):
        ranges = validation.get("range", [None, None])
        return ("Number", {"max_val": ranges[1], "min_val": ranges[0]})

    @classmethod
    def regex(cls, validation):
        return ("Pattern", {"regex": "r" + quote_regex(validation.get("pattern"))})

    @classmethod
    def email(cls, validation):
        regex = (
            r"^[a-zA-Z0-9.!#$%&'*+\/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}"
            r"[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$"
        )
        return ("Pattern", {"regex": "r" + quote_regex(regex)})

    @classmethod
    def ipv4(cls, validation):
        regex = r"^(?:(?:[0-1]?\d{1,2}|2[0-4]\d|25[0-5])(?:\.|$)){4}$"
        return ("Pattern", {"regex": "r" + quote_regex(regex)})

    @classmethod
    def date(cls, validation):
        # iso8601 date time format
        regex = (
            r"^\s*((?:[+-]\d{6}|\d{4})-(?:\d\d-\d\d|W\d\d-\d|W\d\d|\d\d\d|\d\d))"
            r"(?:(T| )(\d\d(?::\d\d(?::\d\d(?:[.,]\d+)?)?)?)([\+\-]\d\d(?::?\d\d)?|\s*Z)?)?$"
        )
        return ("Pattern", {"regex": "r" + quote_regex(regex)})

    @classmethod
    def url(cls, validation):
        regex = (
            r"^(?:(?:https?|ftp|opc\.tcp):\/\/)?(?:\S+(?::\S*)?@)?"
            r"(?:(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
            r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}"
            r"(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|"
            r"(?:(?:[a-z\u00a1-\uffff0-9]+-?_?)*[a-z\u00a1-\uffff0-9]+)"
            r"(?:\.(?:[a-z\u00a1-\uffff0-9]+-?)*[a-z\u00a1-\uffff0-9]+)*"
            r"(?:\.(?:[a-z\u00a1-\uffff]{2,}))?)(?::\d{2,5})?(?:\/[^\s]*)?$"
        )
        return ("Pattern", {"regex": "r" + quote_regex(regex)})

    @classmethod
    def multiple_validators(cls, validators):
        validators_str = ", \n".join(validators)
        _template = """validator.AllOf(\n{validators}\n)"""
        return _template.format(validators=indent(validators_str),)


GlobalConfigValidation.validation_mapping = {
    "string": GlobalConfigValidation.string,
    "number": GlobalConfigValidation.number,
    "regex": GlobalConfigValidation.regex,
    "email": GlobalConfigValidation.email,
    "ipv4": GlobalConfigValidation.ipv4,
    "date": GlobalConfigValidation.date,
    "url": GlobalConfigValidation.url,
}


class GlobalConfigPostProcessor(object):
    """
    Post process for REST builder.
    """

    output_local = "local"
    _import_declare_template = """
import {import_declare_name}
"""

    _import_declare_content = """
import os
import sys
import re
from os.path import dirname

ta_name = '{ta_name}'
pattern = re.compile(r'[\\\\/]etc[\\\\/]apps[\\\\/][^\\\\/]+[\\\\/]bin[\\\\/]?$')
new_paths = [path for path in sys.path if not pattern.search(path) or ta_name in path]
new_paths.append(os.path.join(dirname(dirname(__file__)), "lib"))
new_paths.insert(0, os.path.sep.join([os.path.dirname(__file__), ta_name]))
sys.path = new_paths
"""

    def __init__(self):
        self.builder = None
        self.schema = None
        self.import_declare_name = None

    @property
    def root_path(self):
        return getattr(self.builder.output, "_path")

    def third_path(self):
        return self.schema.namespace

    def default_to_local(self):
        default_dir = op.join(self.root_path, self.builder.output.default,)
        local_dir = op.join(self.root_path, self.output_local,)
        if not op.isdir(local_dir):
            os.makedirs(local_dir)
        for i in os.listdir(default_dir):
            child = op.join(default_dir, i)
            if op.isdir(child):
                shutil.copytree(child, local_dir)
            else:
                shutil.copy(child, op.join(local_dir, i))

        # remove the default folder
        shutil.rmtree(default_dir, ignore_errors=True)

    def import_declare_py_name(self):
        if self.import_declare_name:
            return self.import_declare_name
        return "{}_import_declare".format(self.schema.namespace)

    def import_declare_py_content(self):
        import_declare_file = op.join(
            self.root_path,
            self.builder.output.bin,
            self.import_declare_py_name() + ".py",
        )
        content = self._import_declare_content.format(ta_name=self.schema.namespace,)
        with open(import_declare_file, "w") as f:
            f.write(content)

    def import_declare(self, rh_file):
        with open(rh_file) as f:
            cont = [l for l in f]
        import_declare = self._import_declare_template.format(
            import_declare_name=self.import_declare_py_name()
        )
        cont.insert(0, import_declare)
        with open(rh_file, "w") as f:
            f.write("".join(cont))

    def __call__(self, builder, schema, import_declare_name=None):
        """
        :param builder: REST builder
        :param schema: Global Config Schema
        :return:
        """
        self.builder = builder
        self.schema = schema
        self.import_declare_name = import_declare_name

        self.import_declare_py_content()
        for endpoint in schema.endpoints:
            rh_file = op.join(
                getattr(builder.output, "_path"),
                builder.output.bin,
                endpoint.rh_name + ".py",
            )
            self.import_declare(rh_file)
        # self.default_to_local()

        # add executable permission to files under bin folder
        def add_executable_attribute(file_path):
            if op.isfile(file_path):
                st = os.stat(file_path)
                os.chmod(file_path, st.st_mode | 0o111)

        bin_path = op.join(getattr(builder.output, "_path"), builder.output.bin)
        items = os.listdir(bin_path)
        list(map(add_executable_attribute, items))



from builtins import str
from builtins import object
import requests


class TARestHelper(object):
    def __init__(self, logger=None):
        self.logger = logger
        self.http_session = None
        self.requests_proxy = None

    def _init_request_session(self, proxy_uri=None):
        self.http_session = requests.Session()
        self.http_session.mount(
            'http://', requests.adapters.HTTPAdapter(max_retries=3))
        self.http_session.mount(
            'https://', requests.adapters.HTTPAdapter(max_retries=3))
        if proxy_uri:
            self.requests_proxy = {'http': proxy_uri, 'https': proxy_uri}

    def send_http_request(self, url, method, parameters=None, payload=None, headers=None, cookies=None, verify=True,
                          cert=None, timeout=None, proxy_uri=None):
        if self.http_session is None:
            self._init_request_session(proxy_uri)
        requests_args = {'timeout': (10.0, 5.0), 'verify': verify}
        if parameters:
            requests_args['params'] = parameters
        if payload:
            if isinstance(payload, (dict, list)):
                requests_args['json'] = payload
            else:
                requests_args['data'] = str(payload)
        if headers:
            requests_args['headers'] = headers
        if cookies:
            requests_args['cookies'] = cookies
        if cert:
            requests_args['cert'] = cert
        if timeout is not None:
            requests_args['timeout'] = timeout
        if self.requests_proxy:
            requests_args['proxies'] = self.requests_proxy
        return self.http_session.request(method, url, **requests_args)



"""
Global Config Module
"""

from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import object
import urllib.parse
from solnlib.splunk_rest_client import SplunkRestClient

from .configuration import (
    Inputs,
    Configs,
    Settings,
    GlobalConfigError,
    Configuration
)
from .schema import GlobalConfigSchema


__all__ = [
    'GlobalConfigError',
    'GlobalConfigSchema',
    'GlobalConfig',
    'Inputs',
    'Configs',
    'Settings',
]


class GlobalConfig(object):

    def __init__(self, splunkd_uri, session_key, schema):
        """
        Global Config.

        :param splunkd_uri:
        :param session_key:
        :param schema:
        :type schema: GlobalConfigSchema
        """
        self._splunkd_uri = splunkd_uri
        self._session_key = session_key
        self._schema = schema

        splunkd_info = urllib.parse.urlparse(self._splunkd_uri)
        self._client = SplunkRestClient(
            self._session_key,
            self._schema.product,
            scheme=splunkd_info.scheme,
            host=splunkd_info.hostname,
            port=splunkd_info.port,
        )
        self._configuration = Configuration(self._client, self._schema)
        self._inputs = Inputs(self._client, self._schema)
        self._configs = Configs(self._client, self._schema)
        self._settings = Settings(self._client, self._schema)

    @property
    def inputs(self):
        return self._inputs

    @property
    def configs(self):
        return self._configs

    @property
    def settings(self):
        return self._settings

    # add support for batch save of configuration payload
    def save(self, payload):
        return self._configuration.save(payload)



from __future__ import absolute_import
import os
import os.path as op
from os.path import dirname as dn
from os.path import basename as bn
from shutil import copy
from . import alert_actions_exceptions as aae
from . import arf_consts as ac

from splunk_add_on_ucc_framework.alert_utils.alert_utils_common.conf_parser import TABConfigParser

merge_black_list = ['default.meta', 'README.txt']
merge_mode_config = {
    "app.conf": "item_overwrite"
}


def remove_alert_from_conf_file(alert, conf_file, logger):
    if not alert or not conf_file:
        logger.info('alert="%s", conf_file="%s"', alert, conf_file)
        return

    if not isinstance(alert, dict):
        msg = 'alert="{}", event="alert is not a dict, don\'t remove anything form file {}"'.format(alert, conf_file)
        raise aae.AlertCleaningFormatFailure(msg)

    parser = TABConfigParser()
    parser.read(conf_file)
    conf_dict = parser.item_dict()

    for stanza, key_values in list(conf_dict.items()):
        if stanza == alert[ac.SHORT_NAME] or \
            stanza == alert[ac.SHORT_NAME] + "_modaction_result" or \
                stanza == "eventtype=" + alert[ac.SHORT_NAME] + "_modaction_result":
            logger.info('alert="%s", conf_file="%s", stanza="%s"',
                        alert[ac.SHORT_NAME],
                        conf_file, stanza)
            parser.remove_section(stanza)

    with open(conf_file, "w") as cf:
        parser.write(cf)


def merge_conf_file(src_file, dst_file, merge_mode="stanza_overwrite"):
    if not os.path.isfile(src_file):
        return
    if not os.path.isfile(dst_file):
        return
    if bn(src_file) in merge_black_list:
        return

    sparser = TABConfigParser()
    sparser.read(src_file)
    src_dict = sparser.item_dict()
    parser = TABConfigParser()
    parser.read(dst_file)
    dst_dict = parser.item_dict()

    if merge_mode == "stanza_overwrite":
        for stanza, key_values in list(src_dict.items()):
            if stanza not in dst_dict:
                parser.add_section(stanza)
            else:
                parser.remove_section(stanza,false)
                parser.add_section(stanza)

            for k, v in list(key_values.items()):
                parser.set(stanza, k, v)
    elif merge_mode == "item_overwrite":
        for stanza, key_values in list(src_dict.items()):
            if stanza not in dst_dict:
                parser.add_section(stanza)

            for k, v in list(key_values.items()):
                if v:
                    parser.set(stanza, k, v)
                else:
                    parser.remove_option(stanza, k)
    else:
        # overwrit the whole file
        parser.read(src_file)

    with open(dst_file, "w") as df:
        parser.write(df)


def merge(src, dst, no_black_list=True):
    if op.isfile(src):
        return

    src_files = os.listdir(src)
    dst_files = os.listdir(dst)
    merge_mode = "stanza_overwrite"

    for file in src_files:
        f_path = op.join(src, file)
        if op.isfile(f_path):
            if no_black_list and file in merge_black_list:
                continue

            if file.endswith("pyo") or file.endswith("pyc"):
                continue
            if file in dst_files and (file.endswith('.conf') or file.endswith('.conf.spec')):
                if file in list(merge_mode_config.keys()):
                    merge_mode = merge_mode_config[file]
                merge_conf_file(f_path, op.join(dst, file), merge_mode)
            else:
                copy(f_path, dst)
        elif op.isdir(f_path):
            if file.startswith('.'):
                continue
            if file not in dst_files:
                os.makedirs(op.join(dst, file))
            merge(f_path, op.join(dst, file))
        else:
            raise Exception("Unsupported file type {}".format(f_path))



from __future__ import print_function
from splunk_aoblib.setup_util import Setup_Util
from splunk_aoblib.rest_helper import TARestHelper
import logging
from logging_helper import get_logger
from cim_actions import ModularAction
import requests
from builtins import str
import csv
import gzip
import sys

try:
    from splunk.clilib.bundle_paths import make_splunkhome_path
except ImportError:
    from splunk.appserver.mrsparkle.lib.util import make_splunkhome_path

sys.path.insert(0, make_splunkhome_path(
    ["etc", "apps", "Splunk_SA_CIM", "lib"]))


class ModularAlertBase(ModularAction):
    def __init__(self, ta_name, alert_name):
        self._alert_name = alert_name
        # self._logger_name = "modalert_" + alert_name
        self._logger_name = alert_name + "_modalert"
        self._logger = get_logger(self._logger_name)
        super(ModularAlertBase, self).__init__(
            sys.stdin.read(), self._logger, alert_name)
        self.setup_util_module = None
        self.setup_util = None
        self.result_handle = None
        self.ta_name = ta_name
        self.splunk_uri = self.settings.get('server_uri')
        self.setup_util = Setup_Util(
            self.splunk_uri, self.session_key, self._logger)

        self.rest_helper = TARestHelper(self._logger)

    def log_error(self, msg):
        self.message(msg, 'failure', level=logging.ERROR)

    def log_info(self, msg):
        self.message(msg, 'success', level=logging.INFO)

    def log_debug(self, msg):
        self.message(msg, None, level=logging.DEBUG)

    def log_warn(self, msg):
        self.message(msg, None, level=logging.WARN)

    def set_log_level(self, level):
        self._logger.setLevel(level)

    def get_param(self, param_name):
        return self.configuration.get(param_name)

    def get_global_setting(self, var_name):
        return self.setup_util.get_customized_setting(var_name)

    def get_user_credential(self, username):
        '''
        if the username exists, return
        {
            "username": username,
            "password": credential
        }
        '''
        return self.setup_util.get_credential_by_username(username)

    def get_user_credential_by_account_id(self, account_id):
        '''
        if the account_id exists, return
        {
            "username": username,
            "password": credential
        }
        '''
        return self.setup_util.get_credential_by_id(account_id)

    @property
    def log_level(self):
        return self.get_log_level()

    @property
    def proxy(self):
        return self.get_proxy()

    def get_log_level(self):
        return self.setup_util.get_log_level()

    def get_proxy(self):
        ''' if the proxy setting is set. return a dict like
        {
        proxy_url: ... ,
        proxy_port: ... ,
        proxy_username: ... ,
        proxy_password: ... ,
        proxy_type: ... ,
        proxy_rdns: ...
        }
        '''
        return self.setup_util.get_proxy_settings()

    def _get_proxy_uri(self):
        uri = None
        proxy = self.get_proxy()
        if proxy and proxy.get('proxy_url') and proxy.get('proxy_type'):
            uri = proxy['proxy_url']
            if proxy.get('proxy_port'):
                uri = '{0}:{1}'.format(uri, proxy.get('proxy_port'))
            if proxy.get('proxy_username') and proxy.get('proxy_password'):
                uri = '{0}://{1}:{2}@{3}/'.format(proxy['proxy_type'], proxy[
                                                  'proxy_username'], proxy['proxy_password'], uri)
            else:
                uri = '{0}://{1}'.format(proxy['proxy_type'], uri)
        return uri

    def send_http_request(self, url, method, parameters=None, payload=None, headers=None, cookies=None, verify=True, cert=None, timeout=None, use_proxy=True):
        return self.rest_helper.send_http_request(url=url, method=method, parameters=parameters, payload=payload,
                                                  headers=headers, cookies=cookies, verify=verify, cert=cert,
                                                  timeout=timeout,
                                                  proxy_uri=self._get_proxy_uri() if use_proxy else None)

    def build_http_connection(self, config, timeout=120,
                              disable_ssl_validation=False):
        from httplib2 import (socks, ProxyInfo, Http)
        """
        :config: dict like, proxy and account information are in the following
                format {
                    "username": xx,
                    "password": yy,
                    "proxy_url": zz,
                    "proxy_port": aa,
                    "proxy_username": bb,
                    "proxy_password": cc,
                    "proxy_type": http,http_no_tunnel,sock4,sock5,
                    "proxy_rdns": 0 or 1,
                }
        :return: Http2.Http object
        """
        if not config:
            config = {}

        proxy_type_to_code = {
            "http": socks.PROXY_TYPE_HTTP,
            "http_no_tunnel": socks.PROXY_TYPE_HTTP_NO_TUNNEL,
            "socks4": socks.PROXY_TYPE_SOCKS4,
            "socks5": socks.PROXY_TYPE_SOCKS5,
        }
        if config.get("proxy_type") in proxy_type_to_code:
            proxy_type = proxy_type_to_code[config["proxy_type"]]
        else:
            proxy_type = socks.PROXY_TYPE_HTTP

        rdns = config.get("proxy_rdns")

        proxy_info = None
        if config.get("proxy_url") and config.get("proxy_port"):
            if config.get("proxy_username") and config.get("proxy_password"):
                proxy_info = ProxyInfo(proxy_type=proxy_type,
                                       proxy_host=config["proxy_url"],
                                       proxy_port=int(config["proxy_port"]),
                                       proxy_user=config["proxy_username"],
                                       proxy_pass=config["proxy_password"],
                                       proxy_rdns=rdns)
            else:
                proxy_info = ProxyInfo(proxy_type=proxy_type,
                                       proxy_host=config["proxy_url"],
                                       proxy_port=int(config["proxy_port"]),
                                       proxy_rdns=rdns)
        if proxy_info:
            http = Http(proxy_info=proxy_info, timeout=timeout,
                        disable_ssl_certificate_validation=disable_ssl_validation)
        else:
            http = Http(timeout=timeout,
                        disable_ssl_certificate_validation=disable_ssl_validation)

        if config.get("username") and config.get("password"):
            http.add_credentials(config["username"], config["password"])
        return http

    def process_event(self, *args, **kwargs):
        raise NotImplemented()

    def pre_handle(self, num, result):
        result.setdefault('rid', str(num))
        self.update(result)
        return result

    def get_events(self):
        try:
            self.result_handle = gzip.open(self.results_file, 'rb')
            return (self.pre_handle(num, result) for num, result in enumerate(csv.DictReader(self.result_handle)))
        except IOError:
            msg = "Error: {}."
            self.log_error(msg.format(
                "No search result. Cannot send alert action."))
            sys.exit(2)

    def prepare_meta_for_cam(self):
        with gzip.open(self.results_file, 'rb') as rf:
            for num, result in enumerate(csv.DictReader(rf)):
                result.setdefault('rid', str(num))
                self.update(result)
                self.invoke()
                break

    def run(self, argv):
        status = 0
        if len(argv) < 2 or argv[1] != "--execute":
            msg = 'Error: argv="{}", expected="--execute"'.format(argv)
            print(msg, file=sys.stderr)
            sys.exit(1)

        # prepare meta first for permission lack error handling: TAB-2455
        self.prepare_meta_for_cam()
        try:
            level = self.get_log_level()
            if level:
                self._logger.setLevel(level)
        except Exception as e:
            if str(e) and '403' in str(e):  # Handled e.message with str(e)
                self.log_error('User does not have permissions')
            else:
                self.log_error('Unable to set log level')
            sys.exit(2)

        try:
            status = self.process_event()
        except IOError:
            msg = "Error: {}."
            self.log_error(msg.format(
                "No search result. Cannot send alert action."))
            sys.exit(2)
        except Exception as e:
            msg = "Unexpected error: {}."
            if str(e):  # e.message handled
                self.log_error(msg.format(str(e)))  # e.message handled
            else:
                import traceback
                self.log_error(msg.format(traceback.format_exc()))
            sys.exit(2)
        finally:
            if self.result_handle:
                self.result_handle.close()

        return status




from builtins import object
class BaseHookMixin(object):
    """ Base Hook Mixin class
    """
    def create_hook(self, session_key, config_name, stanza_id, payload):
        """Create hook called before the actual create action

        Args:
            config_name: configuration name
            stanza_id: the id of the stanza to create
            payload: data dict
        """
        pass

    def edit_hook(self, session_key, config_name, stanza_id, payload):
        """Edit hook called before the actual create action

        Args:
            config_name: configuration name
            stanza_id: the id of the stanza to edit
            payload: data dict
        """
        pass

    def delete_hook(self, session_key, config_name, stanza_id):
        """Delete hook called before the actual create action

        Args:
            config_name: configuration name
            stanza_id: the id of the stanza to delete
        """
        pass




from __future__ import absolute_import

from functools import wraps
from splunk import admin
from solnlib.splunkenv import get_splunkd_uri
from solnlib.utils import is_true

from .eai import EAI_FIELDS
from .handler import RestHandler
import os
from .endpoint import (
    SingleModel,
    DataInputModel,
    MultipleModel
)

try:
    from custom_hook_mixin import CustomHookMixin as HookMixin
except ImportError:
    from .base_hook_mixin import BaseHookMixin as HookMixin


__all__ = [
    'make_conf_item',
    'build_conf_info',
    'AdminExternalHandler',
]


def make_conf_item(conf_item, content, eai):
    for key, val in list(content.items()):
        conf_item[key] = val

    for eai_field in EAI_FIELDS:
        conf_item.setMetadata(eai_field, eai.content[eai_field])

    return conf_item


def build_conf_info(meth):
    """
    Build conf info for admin external REST endpoint.

    :param meth:
    :return:
    """
    @wraps(meth)
    def wrapper(self, confInfo):
        result = meth(self, confInfo)
        for entity in result:
            make_conf_item(
                confInfo[entity.name],
                entity.content,
                entity.eai,
            )

    return wrapper

def get_splunkd_endpoint():
    if os.environ.get('SPLUNKD_URI'):
        return os.environ['SPLUNKD_URI']
    else:
        splunkd_uri = get_splunkd_uri()
        os.environ['SPLUNKD_URI'] = splunkd_uri
        return splunkd_uri


class AdminExternalHandler(HookMixin, admin.MConfigHandler, object):

    # Leave it for setting REST model
    endpoint = None

    # action parameter for getting clear credentials
    ACTION_CRED = '--cred--'

    def __init__(self, *args, **kwargs):
        # use classic inheritance to be compatible for
        # old version of Splunk private SDK
        admin.MConfigHandler.__init__(
            self,
            *args,
            **kwargs
        )
        self.handler = RestHandler(
            get_splunkd_endpoint(),
            self.getSessionKey(),
            self.endpoint,
        )
        self.payload = self._convert_payload()

    def setup(self):
        # add args for getting clear credentials
        if self.requestedAction == admin.ACTION_LIST:
            self.supportedArgs.addOptArg(self.ACTION_CRED)

        # add args in payload while creating/updating
        actions = (admin.ACTION_LIST, admin.ACTION_REMOVE)
        if self.requestedAction in actions:
            return
        model = self.endpoint.model(self.callerArgs.id)
        if self.requestedAction == admin.ACTION_CREATE:
            for field in model.fields:
                if field.required:
                    self.supportedArgs.addReqArg(field.name)
                else:
                    self.supportedArgs.addOptArg(field.name)

        if self.requestedAction == admin.ACTION_EDIT:
            for field in model.fields:
                self.supportedArgs.addOptArg(field.name)

    @build_conf_info
    def handleList(self, confInfo):
        decrypt = self.callerArgs.data.get(
            self.ACTION_CRED,
            [False],
        )
        decrypt = is_true(decrypt[0])
        if self.callerArgs.id:
            result = self.handler.get(
                self.callerArgs.id,
                decrypt=decrypt,
            )
        else:
            result = self.handler.all(
                decrypt=decrypt,
                count=0,
            )
        return result

    @build_conf_info
    def handleCreate(self, confInfo):
        self.create_hook(
            session_key=self.getSessionKey(),
            config_name=self._get_name(),
            stanza_id=self.callerArgs.id,
            payload=self.payload
        )
        return self.handler.create(
            self.callerArgs.id,
            self.payload,
        )

    @build_conf_info
    def handleEdit(self, confInfo):
        disabled = self.payload.get('disabled')
        if disabled is None:
            self.edit_hook(
                session_key=self.getSessionKey(),
                config_name=self._get_name(),
                stanza_id=self.callerArgs.id,
                payload=self.payload
            )
            return self.handler.update(
                self.callerArgs.id,
                self.payload,
            )
        elif is_true(disabled):
            return self.handler.disable(self.callerArgs.id)
        else:
            return self.handler.enable(self.callerArgs.id)

    @build_conf_info
    def handleRemove(self, confInfo):
        self.delete_hook(
            session_key=self.getSessionKey(),
            config_name=self._get_name(),
            stanza_id=self.callerArgs.id
        )
        return self.handler.delete(self.callerArgs.id)

    def _get_name(self):
        name = None
        if isinstance(self.handler.get_endpoint(), DataInputModel):
            name = self.handler.get_endpoint().input_type
        elif isinstance(self.handler.get_endpoint(), SingleModel):
            name = self.handler.get_endpoint().config_name
        elif isinstance(self.handler.get_endpoint(), MultipleModel):
            # For multiple model, the configuraiton name is same with stanza id
            name = self.callerArgs.id
        return name

    def _convert_payload(self):
        check_actions = (admin.ACTION_CREATE, admin.ACTION_EDIT)
        if self.requestedAction not in check_actions:
            return None

        payload = {}
        for filed, value in list(self.callerArgs.data.items()):
            payload[filed] = value[0] if value and value[0] else ''
        return payload


def handle(
        endpoint,
        handler=AdminExternalHandler,
        context_info=admin.CONTEXT_APP_ONLY,
):
    """
    Handle request.

    :param endpoint: REST endpoint
    :param handler: REST handler
    :param context_info:
    :return:
    """
    real_handler = type(
        handler.__name__,
        (handler, ),
        {'endpoint': endpoint},
    )
    admin.init(real_handler, ctxInfo=context_info)




from future import standard_library
standard_library.install_aliases()
import json
import traceback
from urllib.parse import urlparse
from solnlib.splunkenv import get_splunkd_uri
from solnlib.splunk_rest_client import SplunkRestClient
from solnlib.conf_manager import ConfManager
from splunktaucclib.rest_handler.error import RestError
from splunktaucclib.rest_handler.admin_external import (
    AdminExternalHandler,
)
from splunktaucclib.rest_handler import util


def _migrate_error_handle(func):
    def handle(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            raise RestError(
                500,
                'Migrating failed. %s' % traceback.format_exc()
            )

    return handle


class ConfigMigrationHandler(AdminExternalHandler):
    """
    REST handler, which will migrate configuration
    from add-on built by previous version of TAB (v2.0.0).
    """

    def handleList(self, confInfo):
        self._migrate()
        # use classic inheritance to be compatible for
        # old version of Splunk private SDK
        AdminExternalHandler.handleList(self, confInfo)

    @_migrate_error_handle
    def _migrate(self):
        internal_endpoint = self.endpoint.internal_endpoint
        if not (internal_endpoint.endswith('settings') or
                internal_endpoint.endswith('account')):
            return

        splunkd_info = urlparse(get_splunkd_uri())
        self.base_app_name = util.get_base_app_name()
        self.conf_mgr = ConfManager(
            self.getSessionKey(),
            self.base_app_name,
            scheme=splunkd_info.scheme,
            host=splunkd_info.hostname,
            port=splunkd_info.port,
        )
        self.client = SplunkRestClient(
            self.getSessionKey(),
            self.base_app_name,
            scheme=splunkd_info.scheme,
            host=splunkd_info.hostname,
            port=splunkd_info.port,
        )
        self.legacy_passwords = None

        # migration legacy configuration in related conf files
        if internal_endpoint.endswith('settings'):
            self._migrate_conf()
            self._migrate_conf_customized()
        elif internal_endpoint.endswith('account'):
            self._migrate_conf_credential()

    def get_legacy_passwords(self):
        if self.legacy_passwords is None:
            self.legacy_passwords = {}
            for pwd in self.client.storage_passwords.list(count=-1):
                if pwd.realm == self.base_app_name:
                    self.legacy_passwords[pwd.username] = pwd
        return self.legacy_passwords

    def _migrate_conf(self):
        """
        Migrate from <TA-name>.conf to <prefix>_settings.conf
        :return:
        """
        if self.callerArgs.id not in ('logging', 'proxy'):
            return
        conf_file_name = self.base_app_name
        conf_file, stanzas = self._load_conf(conf_file_name)
        if not stanzas:
            return

        # migrate: global_settings ==> logging
        if 'global_settings' in stanzas and self.callerArgs.id == 'logging':
            stanza = stanzas['global_settings']
            if 'log_level' in stanza:
                stanza['loglevel'] = stanza['log_level']
                del stanza['log_level']
            name = 'logging'
            response = self.handler.update(
                name,
                self._filter_stanza(name, stanza),
            )
            self._loop_response(response)
            # delete legacy configuration
            self._delete_legacy(conf_file, {'global_settings': None})

        # migrate: proxy_settings ==> proxy
        if 'proxy_settings' in stanzas and self.callerArgs.id == 'proxy':
            name = 'proxy'
            response = self.handler.update(
                name,
                self._filter_stanza(name, stanzas['proxy_settings']),
            )
            self._loop_response(response)
            # delete legacy configuration
            self._delete_legacy(conf_file, {'proxy_settings': None})

    def _migrate_conf_customized(self):
        """
        Migrate from <TA-name>_customized.conf to <prefix>_settings.conf
        :return:
        """
        if self.callerArgs.id != 'additional_parameters':
            return

        conf_file_name = self.base_app_name + '_customized'
        conf_file, stanzas = self._load_conf(conf_file_name)
        if not stanzas:
            return

        additional_parameters = {}
        for stanza_name, stanza in list(stanzas.items()):
            for key, val in list(stanza.items()):
                if key == 'type':
                    continue
                else:
                    additional_parameter = val
                    break
            else:
                continue
            if additional_parameter:
                additional_parameters[stanza_name] = additional_parameter

        name = 'additional_parameters'
        response = self.handler.update(
            name,
            self._filter_stanza(name, additional_parameters),
        )
        self._loop_response(response)

        # delete legacy configuration
        self._delete_legacy(conf_file, stanzas)

    def _migrate_conf_credential(self):
        """
        Migrate from <TA-name>_credential.conf to <prefix>_account.conf
        :return:
        """
        conf_file_name = self.base_app_name + '_credential'
        conf_file, stanzas = self._load_conf(conf_file_name)

        for stanza_name, stanza in list(stanzas.items()):
            stanza['username'] = stanza_name
            response = self.handler.create(
                stanza_name,
                stanza,
            )
            self._loop_response(response)

        # delete legacy configuration
        self._delete_legacy(conf_file, stanzas)

    def _load_conf(self, conf_file_name):
        if conf_file_name not in self.client.confs:
            return None, {}
        conf_file = self.conf_mgr.get_conf(conf_file_name)
        stanzas = conf_file.get_all()
        for stanza_name, stanza in list(stanzas.items()):
            pwd = self.get_legacy_passwords().get(stanza_name)
            if pwd:
                pwd_cont = json.loads(pwd.clear_password)
                stanza.update(pwd_cont)
            for key in list(stanza.keys()):
                if key.startswith('eai:') or key == 'disabled':
                    del stanza[key]

        return conf_file, stanzas

    def _delete_legacy(self, conf_file, stanzas):
        for stanza_name, _ in list(stanzas.items()):
            try:
                # delete stanza from related conf file
                conf_file.delete(stanza_name)
            except Exception:
                pass

            pwd = self.get_legacy_passwords().get(stanza_name)
            try:
                # delete password from passwords.conf
                if pwd:
                    pwd.delete()
            except Exception:
                pass

    def _filter_stanza(self, stanza_name, stanza):
        model = self.endpoint.model(stanza_name, stanza)
        stanza_new = {
            f.name: stanza[f.name] for f in model.fields if f.name in stanza
        }
        return stanza_new

    @classmethod
    def _loop_response(cls, response):
        for _ in response:
            pass



"""Credentials Management for REST Endpoint
"""

from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import object
import json
import urllib.parse
from solnlib.credentials import (
    CredentialManager,
    CredentialNotExistException,
)

from .util import get_base_app_name
from .error import RestError


__all__ = [
    'RestCredentialsContext',
    'RestCredentials',
]


class RestCredentialsContext(object):
    """
    Credentials' context, including realm, username and password.
    """

    REALM = '__REST_CREDENTIAL__#{base_app}#{endpoint}'

    def __init__(self, endpoint, name, *args, **kwargs):
        self._endpoint = endpoint
        self._name = name
        self._args = args
        self._kwargs = kwargs

    def realm(self):
        """
        RestCredentials context ``realm``.
        :return:
        """
        return self.REALM.format(
            base_app=get_base_app_name(),
            endpoint=self._endpoint.internal_endpoint.strip('/'),
        )

    def username(self):
        """
        RestCredentials context ``username``.
        :return:
        """
        return self._name

    def dump(self, data):
        """
        RestCredentials context ``password``.
        Dump data to string.
        :param data: data to be encrypted
        :type data: dict
        :return:
        """
        return json.dumps(data)

    def load(self, string):
        """
        RestCredentials context ``password``.
        Load data from string.
        :param string: data has been decrypted
        :type string: basestring
        :return:
        """
        try:
            return json.loads(string)
        except ValueError:
            raise RestError(
                500,
                'Fail to load encrypted string, invalid JSON'
            )


class RestCredentials(object):
    """
    Credential Management stored in passwords.conf
    """
    # Changed password constant to six '*' to make it consistent with solnlib password constant
    PASSWORD = u'******'
    EMPTY_VALUE = u''

    def __init__(
            self,
            splunkd_uri,
            session_key,
            endpoint
    ):
        self._splunkd_uri = splunkd_uri
        self._splunkd_info = urllib.parse.urlparse(self._splunkd_uri)
        self._session_key = session_key
        self._endpoint = endpoint
        self._realm = '__REST_CREDENTIAL__#{base_app}#{endpoint}'.format(
            base_app=get_base_app_name(),
            endpoint=self._endpoint.internal_endpoint.strip('/')
        )

    def get_encrypted_field_names(self, name):
        return [x.name for x in self._endpoint.model(name).fields if x.encrypted]

    def encrypt_for_create(self, name, data):
        """
            force to encrypt all fields that need to be encrypted
            used for create scenarios
        :param name:
        :param data:
        :return:
        """
        encrypted_field_names = self.get_encrypted_field_names(name)
        encrypting = {}
        for field_name in encrypted_field_names:
            if field_name in data and data[field_name]:
                # if it exist in data and it's not empty,
                # encrypt it and set original value as "****..."
                encrypting[field_name] = data[field_name]
                data[field_name] = self.PASSWORD

        if encrypting:
            # only save credential when the stanza is existing in
            # passwords.conf or encrypting data is not empty
            self._set(name, encrypting)

    def encrypt_for_update(self, name, data):
        """

        :param name:
        :param data:
        :return:
        """
        encrypted_field_names = self.get_encrypted_field_names(name)
        encrypting = {}
        if not encrypted_field_names:
            # return if there are not encrypted fields
            return
        for field_name in encrypted_field_names:
            if field_name in data and data[field_name]:
                if data[field_name] != self.PASSWORD:
                    # if the field in data and not empty and it's not '*******', encrypted it
                    encrypting[field_name] = data[field_name]
                    data[field_name] = self.PASSWORD
                else:
                    # if the field value is '******', keep the original value
                    original_clear_password = self._get(name)
                    if original_clear_password and original_clear_password.get(field_name):
                        encrypting[field_name] = original_clear_password[field_name]
                    else:
                        # original password does not exist, use '******' as password
                        encrypting[field_name] = data[field_name]
            elif field_name in data and not data[field_name]:
                data[field_name] = ''
            else:
                # field not in data
                # if the optional encrypted field is not passed, keep original if it exist
                original_clear_password = self._get(name)
                if original_clear_password and original_clear_password.get(field_name):
                    encrypting[field_name] = original_clear_password[field_name]
                    data[field_name] = self.PASSWORD

        if encrypting:
            self._set(name, encrypting)
        else:
            self.delete(name)

    def decrypt_for_get(self, name, data):
        """
            encrypt password if conf changed and return data that needs to write back to conf
        :param name:
        :param data:
        :return:
        """
        data_need_write_to_conf = dict()
        # password dict needs to be encrypted
        encrypting = dict()
        encrypted_field_names = self.get_encrypted_field_names(name)
        if not encrypted_field_names:
            return
        try:
            # try to get clear password for the entity
            clear_password = self._get(name)
            # password exist for the entity
            for field_name in encrypted_field_names:
                if field_name in data and data[field_name]:
                    if data[field_name] != self.PASSWORD:
                        # if the field exist in data and not equals to '*******'
                        # add to dict to be encrypted, else treat it as unchanged
                        encrypting[field_name] = data[field_name]
                        data_need_write_to_conf[field_name] = self.PASSWORD

                    else:
                        # get clear password for the field
                        data[field_name] = clear_password[field_name]
                        encrypting[field_name] = clear_password[field_name]

            if encrypting and clear_password != encrypting:
                # update passwords.conf if password changed
                self._set(name, encrypting)
        except CredentialNotExistException:
            # password does not exist for the entity
            for field_name in encrypted_field_names:
                if field_name in data and data[field_name]:
                    if data[field_name] != self.PASSWORD:
                        # if the field exist in data and not equals to '*******'
                        # add to dict to be encrypted
                        encrypting[field_name] = data[field_name]
                        data_need_write_to_conf[field_name] = self.PASSWORD
                    else:
                        # treat '*******' as password
                        encrypting[field_name] = self.PASSWORD

            if encrypting:
                # set passwords.conf if encrypting data is not empty
                self._set(name, encrypting)

        return data_need_write_to_conf

    def encrypt(self, name, data):
        """

        :param name:
        :param data:
        :return:
        """
        # Check if encrypt is needed
        model = self._endpoint.model(name)
        need_encrypting = all(field.encrypted for field in model.fields)
        if not need_encrypting:
            return
        try:
            encrypted = self._get(name)
            existing = True
        except CredentialNotExistException:
            encrypted = {}
            existing = False
        encrypting = self._filter(name, data, encrypted)
        self._merge(name, encrypted, encrypting)
        if existing or encrypting:
            # only save credential when the stanza is existing in
            # passwords.conf or encrypting data is not empty
            self._set(name, encrypting)

    def decrypt(self, name, data):
        """

        :param name:
        :param data:
        :return: If the passwords.conf is updated, masked data.
            Else, None.
        """
        try:
            # clear password object loads from json
            encrypted = self._get(name)
            existing = True
        except CredentialNotExistException:
            encrypted = {}
            existing = False
        # get fields to be encrypted
        encrypting = self._filter(name, data, encrypted)
        self._merge(name, encrypted, encrypting)
        if existing or encrypting:
            # only save credential when the stanza is existing in
            # passwords.conf or encrypting data is not empty
            self._set(name, encrypting)
        data.update(encrypting)
        return encrypted

    def decrypt_all(self, data):
        """
        :param data:
        :return: changed stanza list
        """
        credential_manager = CredentialManager(
            self._session_key,
            owner=self._endpoint.user,
            app=self._endpoint.app,
            realm=self._realm,
            scheme=self._splunkd_info.scheme,
            host=self._splunkd_info.hostname,
            port=self._splunkd_info.port
        )

        all_passwords = credential_manager._get_all_passwords()
        # filter by realm
        realm_passwords = [x for x in all_passwords if x['realm'] == self._realm]
        return self._merge_passwords(data, realm_passwords)

    @staticmethod
    def _delete_empty_value_for_dict(dct):
        empty_value_names = [k for k, v in list(dct.items()) if v == '']
        for k in empty_value_names:
            del dct[k]

    def _merge_passwords(self, data, passwords):
        """
            return if some fields need to write with new "******"
        """
        # merge clear passwords to response data
        changed_item_list = []

        password_dict = {pwd['username']: json.loads(pwd['clear_password']) for pwd in passwords}
        # existed passwords models: previously has encrypted value
        existing_encrypted_items = [x for x in data if x['name'] in password_dict]

        # previously has no encrypted value
        not_encrypted_items = [x for x in data if x['name'] not in password_dict]

        # For model that password existed
        # 1.Password changed: Update it and add to changed_item_list
        # 2.Password unchanged: Get the password and update the response data
        for existed_model in existing_encrypted_items:
            name = existed_model['name']
            clear_password = password_dict[name]
            need_write_magic_pwd = False
            need_write_back_pwd = False
            for k, v in list(clear_password.items()):
                # make sure key exist in model content
                if k in existed_model['content']:
                    if existed_model['content'][k] == self.PASSWORD:
                        # set existing as raw value
                        existed_model['content'][k] = v
                    elif existed_model['content'][k] == '':
                        # mark to delete it
                        clear_password[k] = ''
                        need_write_back_pwd = True
                        continue
                    else:
                        need_write_magic_pwd = True
                        need_write_back_pwd = True
                        clear_password[k] = existed_model['content'][k]
                else:
                    # mark to delete it
                    clear_password[k] = ''
                    need_write_back_pwd = True

            # update the password storage
            if need_write_magic_pwd:
                changed_item_list.append(existed_model)

            if need_write_back_pwd:
                self._delete_empty_value_for_dict(clear_password)
                if clear_password:
                    self._set(name, clear_password)
                else:
                    # there's no any pwd any more, directly delete it.
                    self.delete(name)

        # For other models, encrypt the password and return
        for other_model in not_encrypted_items:
            name = other_model['name']
            content = other_model['content']
            encrypted_field_names = self.get_encrypted_field_names(name)
            clear_password = {}
            for field_name in encrypted_field_names:
                # make sure key exist in model content
                if field_name in content and content[field_name] != '':
                    clear_password[field_name] = content[field_name]
            if clear_password:
                self._set(name, clear_password)

        changed_item_list.extend(not_encrypted_items)
        return changed_item_list

    def delete(self, name):
        context = RestCredentialsContext(self._endpoint, name)
        mgr = self._get_manager(context)
        try:
            mgr.delete_password(user=context.username())
        except CredentialNotExistException:
            pass

    def _set(self, name, credentials):
        if credentials is None:
            return
        context = RestCredentialsContext(self._endpoint, name)
        mgr = self._get_manager(context)
        mgr.set_password(
            user=context.username(),
            password=context.dump(credentials)
        )

    def _get(self, name):
        context = RestCredentialsContext(self._endpoint, name)
        mgr = self._get_manager(context)
        try:
            string = mgr.get_password(user=context.username())
        except CredentialNotExistException:
            return None
        return context.load(string)

    def _filter(self, name, data, encrypted_data):
        model = self._endpoint.model(name)
        encrypting_data = {}
        for field in model.fields:
            if not field.encrypted:
                # remove non-encrypted fields
                if field.name in encrypted_data:
                    del encrypted_data[field.name]
                continue
            if field.name not in data:
                # ignore un-posted fields
                continue
            if data[field.name] == self.PASSWORD:
                # ignore already-encrypted fields
                continue
            if data[field.name] != self.EMPTY_VALUE:
                encrypting_data[field.name] = data[field.name]
                # non-empty fields
                data[field.name] = self.PASSWORD
                if field.name in encrypted_data:
                    del encrypted_data[field.name]
        return encrypting_data

    def _merge(self, name, encrypted, encrypting):
        model = self._endpoint.model(name)
        for field in model.fields:
            if field.encrypted is False:
                continue

            val_encrypting = encrypting.get(field.name)
            if val_encrypting:
                encrypted[field.name] = self.PASSWORD
                continue
            elif val_encrypting == self.EMPTY_VALUE:
                del encrypting[field.name]
                encrypted[field.name] = self.EMPTY_VALUE
                continue

            val_encrypted = encrypted.get(field.name)
            if val_encrypted:
                encrypting[field.name] = val_encrypted
                del encrypted[field.name]

    def _get_manager(self, context):
        return CredentialManager(
            self._session_key,
            owner=self._endpoint.user,
            app=self._endpoint.app,
            realm=context.realm(),
            scheme=self._splunkd_info.scheme,
            host=self._splunkd_info.hostname,
            port=self._splunkd_info.port
        )



from __future__ import absolute_import

from builtins import object
from ..util import get_base_app_name
from ..error import RestError


__all__ = [
    'RestModel',
    'RestEndpoint',
    'SingleModel',
    'MultipleModel',
    'DataInputModel',
]


class RestModel(object):

    def __init__(self, fields, name=None):
        """
        REST Model.
        :param name:
        :param fields:
        """
        self.name = name
        self.fields = fields


class RestEndpoint(object):
    """
    REST Endpoint.
    """

    def __init__(
            self,
            user='nobody',
            app=None,
            *args,
            **kwargs
    ):
        """

        :param user:
        :param app: if None, it will be base app name
        :param args:
        :param kwargs:
        """
        self.user = user
        self.app = app or get_base_app_name()
        self.args = args
        self.kwargs = kwargs

        # If reload is needed while GET request
        self.need_reload = False

    @property
    def internal_endpoint(self):
        """
        Endpoint of Splunk internal service.

        :return:
        """
        raise NotImplementedError()

    def model(self, name):
        """
        Real model for given name.

        :param name:
        :return:
        """
        raise NotImplementedError()

    def _loop_fields(self, meth, name, data, *args, **kwargs):
        model = self.model(name)
        return [getattr(f, meth)(data, *args, **kwargs) for f in model.fields]

    def validate(self, name, data, existing=None):
        self._loop_fields('validate', name, data, existing=existing)

    def encode(self, name, data):
        self._loop_fields('encode', name, data)

    def decode(self, name, data):
        self._loop_fields('decode', name, data)


class SingleModel(RestEndpoint):
    """
    REST Model with Single Mode. It will store stanzas
    with same format  into one conf file.
    """

    def __init__(
            self,
            conf_name,
            model,
            user='nobody',
            app=None,
            *args,
            **kwargs
    ):
        """

        :param conf_name: conf file name
        :param model: REST model
        :type model: RestModel
        :param args:
        :param kwargs:
        """
        super(SingleModel, self).__init__(
            user=user, app=app, *args, **kwargs)
        self.need_reload = True

        self._model = model
        self.conf_name = conf_name
        self.config_name = kwargs.get('config_name')

    @property
    def internal_endpoint(self):
        return 'configs/conf-{}'.format(self.conf_name)

    def model(self, name):
        return self._model


class MultipleModel(RestEndpoint):
    """
    REST Model with Multiple Modes. It will store
     stanzas with different formats into one conf file.
    """

    def __init__(
            self,
            conf_name,
            models,
            user='nobody',
            app=None,
            *args,
            **kwargs
    ):
        """

        :param conf_name:
        :type conf_name: basestring
        :param models: list of RestModel
        :type models: list
        :param args:
        :param kwargs:
        """
        super(MultipleModel, self).__init__(
            user=user, app=app, *args, **kwargs)
        self.need_reload = True

        self.conf_name = conf_name
        self.models = {model.name: model for model in models}

    @property
    def internal_endpoint(self):
        return 'configs/conf-{}'.format(self.conf_name)

    def model(self, name):
        try:
            return self.models[name]
        except KeyError:
            raise RestError(404, 'name=%s' % name)


class DataInputModel(RestEndpoint):
    """
    REST Model for Data Input.
    """

    def __init__(
            self,
            input_type,
            model,
            user='nobody',
            app=None,
            *args,
            **kwargs
    ):
        super(DataInputModel, self).__init__(
            user=user, app=app, *args, **kwargs)

        self.input_type = input_type
        self._model = model

    @property
    def internal_endpoint(self):
        return 'data/inputs/{}'.format(self.input_type)

    def model(self, name):
        return self._model




from __future__ import absolute_import

from builtins import object
from ..error import RestError


__all__ = ['RestField']


class RestField(object):
    """
    REST Field.
    """

    def __init__(
            self,
            name,
            required=False,
            encrypted=False,
            default=None,
            validator=None,
            converter=None,
    ):
        self.name = name
        self.required = required
        self.encrypted = encrypted
        self.default = default
        self.validator = validator
        self.converter = converter

    def validate(self, data, existing=None):
        # update case: check required field in data
        if existing and self.name in data and not data.get(self.name) and self.required:
            raise RestError(
                    400,
                    'Required field is missing: %s' % self.name
                )
        value = data.get(self.name)
        if not value and existing is None:
            if self.required:
                raise RestError(
                    400,
                    'Required field is missing: %s' % self.name
                )
            return
        if self.validator is None or not value:
            return

        res = self.validator.validate(value, data)
        if not res:
            raise RestError(400, self.validator.msg)

    def encode(self, data):
        value = data.get(self.name)
        if not value or self.converter is None:
            return
        data[self.name] = self.converter.encode(value, data)

    def decode(self, data):
        value = data.get(self.name)
        if not value or self.converter is None:
            return
        data[self.name] = self.converter.decode(value, data)



"""
Validators for Splunk configuration.
"""

from __future__ import absolute_import

from builtins import str
from past.builtins import basestring
import sys

from builtins import object
import re
import json
from inspect import isfunction


__all__ = [
    'Validator',
    'ValidationError',
    'AnyOf',
    'AllOf',
    'RequiresIf',
    'UserDefined',
    'Enum',
    'Number',
    'String',
    'Pattern',
    'Host',
    'Port',
    'Datetime',
    'Email',
    'JsonString',
]
basestring = str if sys.version_info[0] == 3 else basestring


class Validator(object):
    """
    Base class of validators.
    """

    def __init__(self):
        # Validation error message queue
        self._msgs = []

    def validate(self, value, data):
        """
        Check if the given value is valid. It assumes that
        the given value is a string.

        :param value: value to validate.
        :param data: whole payload in request.
        :return If the value is invalid, return True.
            Or return False.
        """
        raise NotImplementedError(
            'Function "validate" needs to be implemented.'
        )

    @property
    def msg(self):
        """
        It will return the one with highest priority.

        :return:
        """
        return self._msgs[0] if self._msgs else 'Invalid input value'

    def put_msg(self, msg, high_priority=False):
        """
        Put message content into pool.

        :param msg: error message content
        :param high_priority: is this message with high priority
        :return:
        """
        if high_priority:
            self._msgs.insert(0, msg)
        else:
            self._msgs.append(msg)


class ValidationFailed(Exception):
    """
    Validation error.
    """
    pass


class AnyOf(Validator):
    """
    A composite of validators that accepts values accepted by
    any of its component validators.
    """

    def __init__(self, *validators):
        """

        :param validators: A list of validators.
        """
        super(AnyOf, self).__init__()
        self._validators = validators

    def validate(self, value, data):
        msgs = []
        for validator in self._validators:
            if not validator.validate(value, data):
                msgs.append(validator.msg)
            else:
                return True
        else:
            self.put_msg(
                'At least one of the following errors need to be fixed: %s' % json.dumps(msgs)
            )
            return False


class AllOf(Validator):
    """
    A composite of validators that accepts values accepted by
    all of its component validators.
    """

    def __init__(self, *validators):
        """

        :param validators: A list of validators.
        """
        super(AllOf, self).__init__()
        self._validators = validators

    def validate(self, value, data):
        msgs = []
        for validator in self._validators:
            if not validator.validate(value, data):
                msgs.append(validator.msg)
        if msgs:
            self.put_msg(
                'All of the following errors need to be fixed: %s' % json.dumps(msgs)
            )
            return False
        return True


class RequiresIf(Validator):
    """
    If the given field makes the specified condition as True,
    it requires some other fields are not empty
    in the payload of request.
    """

    def __init__(self, fields, condition=None):
        """

        :param fields: conditionally required field name list.
        :param condition: it can be:
            1. None means any non-empty string for given field
            2. A function takes value & data as parameters and
               returns a boolean value
        """
        assert isinstance(fields, (list, set, tuple)), \
            'Argument "fields" should be list, set or tuple'
        super(RequiresIf, self).__init__()
        self.fields = fields
        self.condition = condition

    @classmethod
    def _is_empty(cls, value):
        return value is None or value == ''

    def validate(self, value, data):
        if self.condition is None and not self._is_empty(value):
            need_validate = True
        else:
            assert isfunction(self.condition), \
                'Condition should be a function for RequiresIf validator'
            need_validate = self.condition(value, data)
        if not need_validate:
            return True

        fields = []
        for field in self.fields:
            val = data.get(field)
            if val is None or val == '':
                fields.append(field)
        if fields:
            self.put_msg(
                'For given input, fields are required: %s' % ', '.join(fields)
            )
            return False
        return True


class UserDefined(Validator):
    """
    A validator that defined by user.

    The user-defined validator function should be in form:
    ``def func(value, data, *args, **kwargs): ...``
    ValidationFailed will be raised if validation failed.

    Usage::
    >>> def my_validate(value, data, args):
    >>>     if value != args or not data:
    >>>         raise ValidationFailed('Invalid input')
    >>>
    >>> my_validator = UserDefined(my_validate, 'test_val')
    >>> my_validator.validate('value', {'key': 'value'}, 'value1')

    """

    def __init__(self, validator, *args, **kwargs):
        """

        :param validator: user-defined validating function
        """
        super(UserDefined, self).__init__()
        self._validator = validator
        self._args = args
        self._kwargs = kwargs

    def validate(self, value, data):
        try:
            self._validator(
                value,
                data,
                *self._args,
                **self._kwargs
            )
        except ValidationFailed as exc:
            self.put_msg(str(exc))
            return False
        else:
            return True


class Enum(Validator):
    """
    A validator that accepts only a finite set of values.
    """

    def __init__(self, values=()):
        """

        :param values: The collection of valid values
        """
        super(Enum, self).__init__()
        try:
            self._values = set(values)
        except TypeError:
            self._values = list(values)

        self.put_msg(
            'Value should be in %s' % json.dumps(list(self._values))
        )

    def validate(self, value, data):
        return value in self._values


class Number(Validator):
    """
    A validator that accepts values within a certain range.
    This is for numeric value.

    Accepted condition: min_val <= value <= max_val
    """

    def __init__(self, min_val=None, max_val=None, is_int=False):
        """

        :param min_val: if not None, it requires min_val <= value
        :param max_val: if not None, it requires value < max_val
        :param is_int: the value should be integer or not
        """
        def check(val):
            try:
                return val is None or isinstance(val, (int, float))
            except NameError:
                return val is None or isinstance(val, (int, float))
        assert check(min_val) and check(max_val), \
            '%(min_val)s & %(max_val)s should be numbers' % {
                'min_val': min_val,
                'max_val': max_val
            }

        super(Number, self).__init__()
        self._min_val = min_val
        self._max_val = max_val
        self._is_int = is_int

    def validate(self, value, data):
        try:
            try:
                value = int(value) if self._is_int else float(value)
            except NameError:
                value = int(value) if self._is_int else float(value)
        except ValueError:
            self.put_msg(
                'Invalid format for %s value' % ('integer' if self._is_int else 'numeric')
            )
            return False

        msg = None
        if not self._min_val and self._max_val and value > self._max_val:
            msg = 'Value should be smaller than %(max_val)s' % {
                'max_val': self._max_val
            }
        elif not self._max_val and self._min_val and value < self._min_val:
            msg = 'Value should be no smaller than %(min_val)s' % {
                'min_val': self._min_val
            }
        elif self._min_val and self._max_val:
            if value < self._min_val or value > self._max_val:
                msg = 'Value should be between %(min_val)s and %(max_val)s' % {
                    'min_val': self._min_val,
                    'max_val': self._max_val
                }
        if msg is not None:
            self.put_msg(msg)
            return False
        return True


class String(Validator):
    """
    A validator that accepts string values.

    Accepted condition: min_len <= len(value) < max_len
    """

    def __init__(self, min_len=None, max_len=None):
        """

        :param min_len: If not None,
            it should be shorter than ``min_len``
        :param max_len: If not None,
            it should be longer than ``max_len``
        """

        def check(val):
            if val is None:
                return True
            try:
                return isinstance(val, int) and val >= 0
            except NameError:
                return isinstance(val, (int)) and val >= 0

        assert check(min_len) and check(max_len), \
            '%(min_len)s & %(max_len)s should be numbers' % {
                'min_len': min_len,
                'max_len': max_len
            }

        super(String, self).__init__()
        self._min_len, self._max_len = min_len, max_len

    def validate(self, value, data):
        if not isinstance(value, basestring):
            self.put_msg('Input value should be string')
            return False

        str_len = len(value)
        msg = None

        if not self._min_len and self._max_len and str_len > self._max_len:
            msg = 'String should be shorter than %(max_len)s' % {
                'max_len': self._max_len
            }
        elif self._min_len and not self._max_len and str_len < self._min_len:
            msg = 'String should be no shorter than %(min_len)s' % {
                'min_len': self._min_len
            }
        elif self._min_len and self._max_len:
            if str_len < self._min_len or str_len > self._max_len:
                msg = 'String length should be between %(min_len)s and %(max_len)s' % {
                    'min_len': self._min_len,
                    'max_len': self._max_len
                }
        if msg is not None:
            self.put_msg(msg)
            return False
        return True


class Datetime(Validator):
    """
    Date time validation.
    """
    def __init__(self, datetime_format):
        """

        :param datetime_format: Date time format,
            e.g. %Y-%m-%dT%H:%M:%S.%f
        """
        super(Datetime, self).__init__()
        self._format = datetime_format

    def validate(self, value, data):
        import datetime
        try:
            datetime.datetime.strptime(value, self._format)
        except ValueError as exc:
            error = 'Wrong datetime with format "%s": %s' % (self._format, str(exc))
            self.put_msg(error)
            return False
        return True


class Pattern(Validator):
    """
    A validator that accepts strings that match
    a given regular expression.
    """

    def __init__(self, regex, flags=0):
        """

        :param regex: The regular expression (string or compiled)
            to be matched.
        :param flags: flags value for regular expression.
        """
        super(Pattern, self).__init__()
        self._regexp = re.compile(regex, flags=flags)
        self.put_msg('Not matching the pattern: %s' % regex)

    def validate(self, value, data):
        return self._regexp.match(value) and True or False


class Host(Pattern):
    """
    A validator that accepts strings that represent network hostname.
    """
    def __init__(self):
        regexp = (
            r'^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*'
            r'([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$'
        )
        super(Host, self).__init__(regexp, flags=re.I)
        self.put_msg('Invalid hostname', high_priority=True)


class Port(Number):
    """
    Port number.
    """
    def __init__(self):
        super(Port, self).__init__(
            min_val=0,
            max_val=65536,
            is_int=True,
        )
        self.put_msg(
            'Invalid port number, it should be a integer between 0 and 65535',
            high_priority=True,
        )


class Email(Pattern):
    """
    A validator that accepts strings that represent network hostname.
    """

    def __init__(self):
        regexp = (
            r'^[A-Z0-9][A-Z0-9._%+-]{0,63}@'
            r'(?:[A-Z0-9](?:[A-Z0-9-]{0,62}[A-Z0-9])?\.){1,8}[A-Z]{2,63}$'
        )
        super(Email, self).__init__(regexp, flags=re.I)
        self.put_msg('Invalid email address', high_priority=True)


class JsonString(Validator):
    """
    Check if the given value is valid JSON string.
    """

    def validate(self, value, data):
        try:
            json.loads(value)
        except ValueError:
            self.put_msg('Invalid JSON string')
            return False
        return True




from __future__ import absolute_import

from builtins import object
from .eai import RestEAI

__all__ = ['RestEntity']


class RestEntity(object):

    def __init__(self, name, content, model, user, app, acl=None):
        self.name = name
        self.content = content
        self.model = model
        self._eai = RestEAI(self.model, user, app, acl)

    @property
    def eai(self):
        return self._eai



"""
REST Handler.
"""

from __future__ import absolute_import

from builtins import str
from future import standard_library
standard_library.install_aliases()
from builtins import object
import json
import traceback
import urllib.parse
from functools import wraps
from solnlib.packages.splunklib import binding
from solnlib.splunk_rest_client import SplunkRestClient

from .error import RestError
from .entity import RestEntity
from .credentials import RestCredentials


__all__ = ['RestHandler']


def _check_name_for_create(name):
    if name == 'default':
        raise RestError(
            400,
            '"%s" is not allowed for entity name' % name
        )
    if name.startswith("_"):
        raise RestError(
            400,
            'Name starting with "_" is not allowed for entity'
        )


def _pre_request(existing):
    """
    Encode payload before request.
    :param existing:
        if True: means must exist
        if False: means must NOT exist
    :return:
    """

    def _pre_request_wrapper(meth):
        """

        :param meth: RestHandler instance method
        :return:
        """
        def check_existing(self, name):
            if not existing:
                # for create, check name
                _check_name_for_create(name)
            # check if the entity existed
            entities = []
            try:
                entities = list(self.get(name))
            except RestError:
                pass

            if existing and not entities:
                raise RestError(
                    404,
                    '"%s" does not exist' % name,
                )
            elif not existing and entities:
                raise RestError(
                    409,
                    'Name "%s" is already in use' % name,
                )

            if entities:
                return entities[0].content
            else:
                return None

        @wraps(meth)
        def wrapper(self, name, data):
            self._endpoint.validate(
                name,
                data,
                check_existing(self, name),
            )
            self._endpoint.encode(name, data)

            return meth(self, name, data)

        return wrapper

    return _pre_request_wrapper


def _decode_response(meth):
    """
    Decode response body.
    :param meth: RestHandler instance method
    :return:
    """
    def decode(self, name, data, acl):
        self._endpoint.decode(name, data)
        return RestEntity(
            name,
            data,
            self._endpoint.model(name),
            self._endpoint.user,
            self._endpoint.app,
            acl=acl,
        )

    @wraps(meth)
    def wrapper(self, *args, **kwargs):
        try:
            for name, data, acl in meth(self, *args, **kwargs):
                yield decode(self, name, data, acl)
        except RestError:
            raise
        except binding.HTTPError as exc:
            raise RestError(exc.status, str(exc))
        except Exception:
            raise RestError(500, traceback.format_exc())

    return wrapper


class RestHandler(object):
    def __init__(
            self,
            splunkd_uri,
            session_key,
            endpoint,
            *args,
            **kwargs
    ):
        self._splunkd_uri = splunkd_uri
        self._session_key = session_key
        self._endpoint = endpoint
        self._args = args
        self._kwargs = kwargs

        splunkd_info = urllib.parse.urlparse(self._splunkd_uri)
        self._client = SplunkRestClient(
            self._session_key,
            self._endpoint.app,
            scheme=splunkd_info.scheme,
            host=splunkd_info.hostname,
            port=splunkd_info.port,
        )
        self.rest_credentials = RestCredentials(
            self._splunkd_uri,
            self._session_key,
            self._endpoint,
        )
        self.PASSWORD = u'******'

    @_decode_response
    def get(self, name, decrypt=False):
        if self._endpoint.need_reload:
            self.reload()
        response = self._client.get(
            self.path_segment(
                self._endpoint.internal_endpoint,
                name=name,
            ),
            output_mode='json',
        )
        return self._format_response(response, get=True, decrypt=decrypt)

    @_decode_response
    def all(self, decrypt=False, **query):
        if self._endpoint.need_reload:
            self.reload()
        response = self._client.get(
            self.path_segment(self._endpoint.internal_endpoint),
            output_mode='json',
            **query
        )
        return self._format_all_response(response, decrypt)

    def get_encrypted_field_names(self, name):
        return [x.name for x in self._endpoint.model(name).fields if x.encrypted]

    @_decode_response
    @_pre_request(existing=False)
    def create(self, name, data):
        data['name'] = name
        self.rest_credentials.encrypt_for_create(name, data)
        response = self._client.post(
            self.path_segment(self._endpoint.internal_endpoint),
            output_mode='json',
            body=data
        )
        return self._format_response(response)

    @_decode_response
    @_pre_request(existing=True)
    def update(self, name, data):
        self.rest_credentials.encrypt_for_update(name, data)
        response = self._client.post(
            self.path_segment(
                self._endpoint.internal_endpoint,
                name=name,
            ),
            output_mode='json',
            body=data
        )
        return self._format_response(response)

    @_decode_response
    def delete(self, name):
        response = self._client.delete(
            self.path_segment(
                self._endpoint.internal_endpoint,
                name=name,
            ),
            output_mode='json',
        )
        # delete credentials if there are encrypted fields
        if self.get_encrypted_field_names(name):
            rest_credentials = RestCredentials(
                self._splunkd_uri,
                self._session_key,
                self._endpoint,
            )
            rest_credentials.delete(name)
        return self._flay_response(response)

    @_decode_response
    def disable(self, name):
        response = self._client.post(
            self.path_segment(
                self._endpoint.internal_endpoint,
                name=name,
                action='disable',
            ),
            output_mode='json',
        )
        return self._flay_response(response)

    @_decode_response
    def enable(self, name):
        response = self._client.post(
            self.path_segment(
                self._endpoint.internal_endpoint,
                name=name,
                action='enable',
            ),
            output_mode='json',
        )
        return self._flay_response(response)

    def reload(self):
        self._client.get(
            self.path_segment(
                self._endpoint.internal_endpoint,
                action='_reload',
            ),
        )

    def get_endpoint(self):
        return self._endpoint

    @classmethod
    def path_segment(cls, endpoint, name=None, action=None):
        """
        Make path segment for given context in Splunk REST format:
        <endpoint>/<entity>/<action>

        :param endpoint: Splunk REST endpoint, e.g. data/inputs
        :param name: entity name for request, "/" will be quoted
        :param action: Splunk REST action, e.g. disable, enable
        :return:
        """
        template = '{endpoint}{entity}{action}'
        entity = ''
        if name:
            # all special characters except "/" will be
            # url-encoded in splunklib.binding.UrlEncoded
            entity = '/' + name.replace('/', '%2F')
        path = template.format(
            endpoint=endpoint.strip('/'),
            entity=entity,
            action='/%s' % action if action else '',
        )
        return path.strip('/')

    def _format_response(self, response, get=False, decrypt=False):
        body = response.body.read()
        try:
            cont = json.loads(body)
        except ValueError:
            raise RestError(
                500,
                'Fail to load response, invalid JSON'
            )
        for entry in cont['entry']:
            name = entry['name']
            data = entry['content']
            acl = entry['acl']
            encrypted_field_names = self.get_encrypted_field_names(name)
            # encrypt and get clear password for get request
            if get:
                masked = self.rest_credentials.decrypt_for_get(name, data)
                if masked:
                    self._client.post(
                        self.path_segment(
                            self._endpoint.internal_endpoint,
                            name=name,
                        ),
                        body=masked
                    )

            if not decrypt:
                # replace clear password with '******'
                for field_name in encrypted_field_names:
                    if field_name in data and data[field_name]:
                        data[field_name] = self.PASSWORD

            yield name, data, acl

    def _flay_response(self, response, decrypt=False):
        body = response.body.read()
        try:
            cont = json.loads(body)
        except ValueError:
            raise RestError(
                500,
                'Fail to load response, invalid JSON'
            )
        for entry in cont['entry']:
            name = entry['name']
            data = entry['content']
            acl = entry['acl']
            if self._need_decrypt(name, data, decrypt):
                self._load_credentials(name, data)
            if not decrypt:
                self._clean_credentials(name, data)
            yield name, data, acl

    def _format_all_response(self, response, decrypt=False):
        body = response.body.read()
        try:
            cont = json.loads(body)
        except ValueError:
            raise RestError(
                500,
                'Fail to load response, invalid JSON'
            )
        # cont['entry']: collection list, load credentials in one request
        if self.get_encrypted_field_names(None):
            self._encrypt_raw_credentials(cont['entry'])
        if not decrypt:
            self._clean_all_credentials(cont['entry'])

        for entry in cont['entry']:
            name = entry['name']
            data = entry['content']
            acl = entry['acl']
            yield name, data, acl

    def _load_credentials(self, name, data):
        rest_credentials = RestCredentials(
            self._splunkd_uri,
            self._session_key,
            self._endpoint
        )
        masked = rest_credentials.decrypt(name, data)
        if masked:
            # passwords.conf changed
            self._client.post(
                self.path_segment(
                    self._endpoint.internal_endpoint,
                    name=name,
                ),
                **masked
            )

    def _encrypt_raw_credentials(self, data):
        rest_credentials = RestCredentials(
            self._splunkd_uri,
            self._session_key,
            self._endpoint
        )
        # get clear passwords for response data and get the password change list
        change_list = rest_credentials.decrypt_all(data)

        field_names = self.get_encrypted_field_names(None)
        for model in change_list:
            # only updates the defined fields in schema
            masked = dict()
            for field in field_names:
                if field in model['content'] and \
                        model['content'][field] != '' and \
                        model['content'][field] != self.PASSWORD:
                    masked[field] = self.PASSWORD

            if masked:
                self._client.post(
                    self.path_segment(
                        self._endpoint.internal_endpoint,
                        name=model['name'],
                    ),
                    body=masked
                )

    def _need_decrypt(self, name, data, decrypt):
        # some encrypted-needed fields are plain text in *.conf.
        encrypted_field = False
        for field in self._endpoint.model(name).fields:
            if field.encrypted is False:
                # ignore non-encrypted fields
                continue
            encrypted_field = True
            if not data.get(field.name):
                # ignore un-stored/empty fields
                continue
            if data[field.name] == RestCredentials.PASSWORD:
                # ignore already-encrypted fields
                continue
            return True

        if decrypt and encrypted_field:
            # clear credentials is required by request and
            # there are some encrypted-needed fields
            return True
        return False

    def _clean_credentials(self, name, data):
        encrypted_field_names = self.get_encrypted_field_names(name)
        for field_name in encrypted_field_names:
            if field_name in data:
                del data[field_name]

    def _clean_all_credentials(self, data):
        encrypted_field_names = self.get_encrypted_field_names(None)
        for model in data:
            for field_name in encrypted_field_names:
                if field_name in model['content'] and \
                        model['content'][field_name] != '':
                    model['content'][field_name] = self.PASSWORD




from __future__ import absolute_import

from .single_model import RestEndpointBuilder, RestEntityBuilder


class DataInputEntityBuilder(RestEntityBuilder):
    def __init__(self, name, fields, input_type, **kwargs):
        super(DataInputEntityBuilder, self).__init__(name, fields, **kwargs)
        self._input_type = input_type

    @property
    def name_spec(self):
        return "{}://<name>".format(self._input_type)

    @property
    def name_default(self):
        return self._input_type

    @property
    def name_rh(self):
        return ""


class DataInputEndpointBuilder(RestEndpointBuilder):

    _rh_template = """
from splunktaucclib.rest_handler.endpoint import (
    field,
    validator,
    RestModel,
    DataInputModel,
)
from splunktaucclib.rest_handler import admin_external, util
from {handler_module} import {handler_name}
import logging

util.remove_http_proxy_env_vars()

{entity}


endpoint = DataInputModel(
    '{input_type}',
    model,
)


if __name__ == '__main__':
    logging.getLogger().addHandler(logging.NullHandler())
    admin_external.handle(
        endpoint,
        handler={handler_name},
    )
"""

    def __init__(self, name, namespace, input_type, **kwargs):
        super(DataInputEndpointBuilder, self).__init__(name, namespace, **kwargs)
        self.input_type = input_type

    @property
    def conf_name(self):
        return "inputs"

    def actions(self):
        return ["edit", "list", "remove", "create"]

    def generate_rh(self, handler):
        entity = self._entities[0]
        return self._rh_template.format(
            handler_module=handler.module,
            handler_name=handler.name,
            entity=entity.generate_rh(),
            input_type=self.input_type,
        )



from past.builtins import basestring
from builtins import str
from builtins import next
from builtins import range
from six import string_types as basestring
import six
from builtins import object
import collections
import csv
import json
import logging
import logging.handlers
import os
import random
import re
import splunk.rest as rest
import time

from splunk.clilib.bundle_paths import make_splunkhome_path
from splunk.util                import mktimegm, normalizeBoolean

# set the maximum allowable CSV field size
#
# The default of the csv module is 128KB; upping to 10MB. See SPL-12117 for
# the background on issues surrounding field sizes.
# (this method is new in python 2.5)
csv.field_size_limit(10485760)


class InvalidResultID(Exception):
    pass


class ModularAction(object):
    DEFAULT_MSGFIELDS = ['signature',
                         'action_name',
                         'search_name',
                         'sid',
                         'orig_sid',
                         'rid',
                         'orig_rid',
                         'app',
                         'user',
                         'action_mode',
                         'action_status']
    DEFAULT_MESSAGE   = 'sendmodaction - ' + ' '.join(['{i}="{{d[{i}]}}"'.format(i=i) for i in DEFAULT_MSGFIELDS])
    # The above yields a string.format() compatible format string:
    #
    #   'sendmodaction - signature="{d[signature]}" action_name="{d[action_name]}"
    #   search_name="{d[search_name]}" sid="{d[sid]}" orig_sid="{d[orig_sid]}"
    #   rid="{d[rid]}" orig_rid="{d[orig_rid]}" app="{d[app]}" user="{d[user]}"
    #   action_mode="{d[action_mode]}" action_status="{d[action_status]}"'

    DEFAULT_DROPEXP  = lambda x: ((x.startswith('_') and x not in ['_raw', '_time'])
                                  or x.startswith('date_')
                                  or x in ['punct', 'sid', 'rid', 'orig_sid', 'orig_rid'])

    DEFAULT_MAPEXP   = lambda x: (x.startswith('tag::')
                                  or x in ['_time', '_raw', 'splunk_server', 'index',
                                           'source', 'sourcetype', 'host', 'linecount',
                                           'timestartpos', 'timeendpos', 'eventtype',
                                           'tag', 'search_name', 'event_hash', 'event_id'])

    DEFAULT_HEADER   = '***SPLUNK*** index="%s" host="%s" source="%s"'
    DEFAULT_BREAKER  = '==##~~##~~  1E8N3D4E6V5E7N2T9 ~~##~~##==\n'
    DEFAULT_IDLINE   = '***Common Action Model*** orig_action_name="%s" orig_sid="%s" orig_rid="%s" sourcetype="%s"\n'
    DEFAULT_INDEX    = 'summary'
    DEFAULT_CHUNK    = 50000

    SHORT_FORMAT     = '%(asctime)s %(levelname)s %(message)s'

    def __init__(self, settings, logger, action_name='unknown'):
        """ Initialize ModularAction class.

        @param settings:    A modular action payload in JSON format.
        @param logger:      A logging instance.
                            Recommend using ModularAction.setup_logger.
        @param action_name: The action name.
                            action_name in payload will take precedence.
        """
        self.settings      = json.loads(settings)
        self.logger        = logger
        self.session_key   = self.settings.get('session_key')
        self.sid           = self.settings.get('sid')
        self.sid_snapshot  = ''
        ## if sid contains rt_scheduler with snapshot-sid; drop snapshot-sid
        ## sometimes self.sid may be an integer (1465593470.1228)
        try:
            rtsid    = re.match('^(rt_scheduler.*)\.(\d+)$', self.sid)
            if rtsid:
                self.sid          = rtsid.group(1)
                self.sid_snapshot = rtsid.group(2)
        except:
            pass

        ## rid_ntuple is a named tuple that represents
        ## the three variables that change on a per-result basis
        self.rid_ntuple    = collections.namedtuple('ID', ['orig_sid','rid','orig_rid'])
        ## rids is a list of rid_ntuple values
        ## automatically maintained by update() calls
        self.rids          = []
        ## current orig_sid based on update()
        ## aka self.rids[-1].orig_sid
        self.orig_sid      = ''
        ## current rid based on update()
        ## aka self.rids[-1].rid
        self.rid           = ''
        ## current orig_rid based on update()
        ## aka self.rids[-1].orig_rid
        self.orig_rid      = ''

        self.results_file  = self.settings.get('results_file')
        ## info
        self.info          = {}
        if self.results_file:
            self.info_file = os.path.join(os.path.dirname(self.results_file), 'info.csv')
        self.search_name   = self.settings.get('search_name')
        self.app           = self.settings.get('app')
        self.user          = self.settings.get('user') or self.settings.get('owner')
        self.configuration = self.settings.get('configuration', {})
        ## enforce configuration is a 'dict'
        if not isinstance(self.configuration, dict):
            self.configuration = {}
        ## set loglevel to DEBUG if verbose
        if normalizeBoolean(self.configuration.get('verbose', 'false')):
            self.logger.setLevel(logging.DEBUG)
            self.logger.debug('loglevel set to DEBUG')
        ## use | sendalert param.action_name=$action_name$
        self.action_name = self.configuration.get('action_name') or action_name
        ## use sid to determine action_mode
        if isinstance(self.sid, basestring) and 'scheduler' in self.sid:
            self.action_mode = 'saved'
        else:
            self.action_mode = 'adhoc'

        self.action_status = ''
        ## Since we don't use the result object we get from settings it will be purged
        try:
            del self.settings['result']
        except Exception:
            pass
        ## events
        self.events = []

    def addinfo(self):
        """ The purpose of this method is to populate the
        modular action info variable with the contents of info.csv.

        @raise Exception: raises Exception if self.info_file could not be opened
                          or if there were problems parsing the info.csv data
        """
        if self.info_file:
            try:
                with open(self.info_file, 'rU') as fh:
                    self.info = next(csv.DictReader(fh))
            except Exception as e:
                self.message('Could not retrieve info.csv', level=logging.WARN)

    def addjobinfo(self):
        """ The purpose of this method is to populate the job variable
        with the contents from REST (/services/search/jobs/<sid>)

        SPL-112815 - sendalert - not all $job.<param>$ parameters come through

        @raise Exception: raises Exception if search job information could not
                          be retrieved via REST (search/jobs) based on self.sid
        """
        self.job = {}
        if self.sid:
            try:
                response, content = rest.simpleRequest('search/jobs/%s' % self.sid,
                                        sessionKey=self.session_key,
                                        getargs={'output_mode': 'json'})
                if response.status == 200:
                    self.job = json.loads(content)['entry'][0]['content']
                    self.message('Successfully retrieved search job info')
                    self.logger.debug(self.job)
                else:
                    self.message('Could not retrieve search job info', level=logging.WARN)
            except Exception as e:
                self.message('Could not retrieve search job info', level=logging.WARN)

    def message(self, signature, status=None, rids=None, level=logging.INFO, **kwargs):
        """ The purpose of this method is to provide a common messaging interface.

        @param signature: A string representing the message we want to log.
        @param status:    An optional status that we want to log.
                          Defaults to None.
        @param rids:      An optional list of rid_ntuple values in case we
                          want to generate the message for multiple rids.
                          Defaults to None (use the rid currently loaded).
        @param level:     The logging level to use when writing the message.
                          Defaults to logging.INFO (INFO)
        @param kwargs:    Additional keyword arguments to be included with the
                          message.
                          Defaults to "no arguments".

        @return message:  This method logs the message; however, for
                          backwards compatibility we also return the message.
        """
        ## status
        status = status or self.action_status or ''
        ## rid
        if not isinstance(rids, list):
            rids = [self.rid_ntuple(self.orig_sid, self.rid, self.orig_rid)]
        ## kwargs - prune any duplicate keys based on DEFAULT_MSGFIELDS
        ##          prune any keys with special characters [A-Za-z_]+
        newargs = [x for x in kwargs\
            if (x not in ModularAction.DEFAULT_MSGFIELDS) and re.match('[A-Za-z_]+', x)]
        ## MSG
        msg = '%s %s' % (ModularAction.DEFAULT_MESSAGE, ' '.join(['{i}="{{d[{i}]}}"'.format(i=i) for i in newargs]))

        # This will set the default value of any value NOT in the dictionary to the
        # empty string.
        argsdict = collections.defaultdict(str)
        # order is important here - here we update first from kwargs, then from our
        # expected arg set.
        argsdict.update(kwargs)
        argsdict.update({
            'signature':     signature or '',
            'action_name':   self.action_name or '',
            'search_name':   self.search_name or '',
            'sid':           self.sid or '',
            'app':           self.app or '',
            'user':          self.user or '',
            'action_mode':   self.action_mode or '',
            'action_status': status
        })

        for rid_ntuple in rids:
            if len(rid_ntuple)==3:
                ## Update the arguments dictionary
                argsdict.update({
                    'orig_sid': rid_ntuple.orig_sid or '',
                    'rid':      rid_ntuple.rid or '',
                    'orig_rid': rid_ntuple.orig_rid or ''
                })
                ## This is where the magic happens. The format string will use the
                ## attributes of "argsdict"
                message = msg.format(d=argsdict)
                ## prune empty string key-value pairs
                for match in re.finditer('[A-Za-z_]+=\"\"(\s|$)', message):
                    message = message.replace(match.group(0),'',1)
                message = message.strip()
                self.logger.log(level, message)
            else:
                self.logger.warn('Could not unpack rid_ntuple')
                message = ''

        return message

    def update(self, result):
        """ The purpose of this method is to update the ModularAction instance
        identifiers based on the current result being operated on.

        This is the most important method in the library as it sets up
        rid, orig_sid, and orig_rid to be used by subsequent class methods.

        Not calling update() immediately for each result before doing additional
        work can have adverse affects.

        @param signature: A string representing the message we want to log.
        @param status:    An optional status that we want to log.
                          Defaults to None.
        @param rids:      An optional list of rid_ntuple values in case we
                          want to generate the message for multiple rids.
                          Defaults to None (use the rid currently loaded).
        @param level:     The logging level to use when writing the message.
                          Defaults to logging.INFO (INFO)
        @param kwargs:    Additional keyword arguments to be included with the
                          message.
                          Defaults to "no arguments".

        @return message:  This method logs the message; however, for
                          backwards compatiblity we also return the message.
        """
        ## This is for events/results that were created as the result of a previous action
        self.orig_sid = result.get('orig_sid', '')
        ## This is for events/results that were created as the result of a previous action
        self.orig_rid = result.get('orig_rid', '')
        if 'rid' in result and isinstance(result['rid'], (basestring, int)):
            self.rid = str(result['rid'])
            if self.sid_snapshot:
                self.rid = '%s.%s' % (self.rid, self.sid_snapshot)
            ## add result info to list of named tuples
            self.rids.append(self.rid_ntuple(self.orig_sid, self.rid, self.orig_rid))
        else:
            raise InvalidResultID('Result must have an ID')

    def invoke(self):
        """ The purpose of this method is to generate per-result invocation messages.
        This method is used to identify that an action is being attempted on a per-result basis.

        Remember to call update() prior to invoke() to ensure that the invocation message
        reflects the appropriate identifiers.
        """
        self.message('Invoking modular action')

    def result2stash(self, result, dropexp=DEFAULT_DROPEXP, mapexp=DEFAULT_MAPEXP, addinfo=False):
        """ The purpose of this method is to formulate an event in stash format

        @param result:  The result dictionary to generate a stash event for.
        @param dropexp: A lambda expression used to determine whether a field
                        should be dropped or not.
                        Defaults to DEFAULT_DROPEXP.
        @param mapexp:  A lambda expression used to determine whether a field
                        should be mapped (prepended with "orig_") or not.
                        Defaults to DEFAULT_MAPEXP.
        @param addinfo: Whether or not to add search information to the event.
                        "info" includes search_now, info_min_time, info_max_time,
                        and info_search_time fields.
                        Requires that information was loaded into the ModularAction
                        instance via addinfo()

        @return _raw:   Returns a string which represents the result in stash format.

        The following example has been broken onto multiple lines for readability:
        06/21/2016 10:00:00 -0700,
        search_name="Access - Brute Force Access Behavior Detected - Rule",
        search_now=0.000, info_min_time=1466528400.000, info_max_time=1466532600.000, info_search_time=1465296264.179,
        key1=key1val, key2=key2val, key3=key3val, key4=key4val1, key4=key4val2, ...
        """
        dropexp      = dropexp or (lambda x: False)
        mapexp       = mapexp  or (lambda x: False)
        orig_dropexp = lambda x: x.startswith('orig_') and x[5:] in result and mapexp(x[5:])

        ## addinfo
        if addinfo:
            result['info_min_time']    = self.info.get('_search_et', '0.000')
            info_max_time              = self.info.get('_search_lt')
            if not info_max_time or info_max_time==0 or info_max_time=='0':
                info_max_time = '+Infinity'
            result['info_max_time']    = info_max_time
            result['info_search_time'] = self.info.get('_timestamp', '')

        ## construct _raw
        _raw = '%s' % result.get('_time', mktimegm(time.gmtime()))
        if self.search_name:
            _raw += ', search_name="%s"' % self.search_name

        processed_keys = []
        for key, val in sorted(result.items()):
            vals = []
            ## if we have a proper mv field
            if (key.startswith('__mv_')
                and val and isinstance(val, basestring)
                and val.startswith('$') and val.endswith('$')):
                real_key = key[5:]
                vals     = val[1:-1].split('$;$')
            ## if proper sv field
            elif val and not key.startswith('__mv_'):
                real_key = key
                vals     = [val]

            ## if we have vals and key hasn't been processed
            ## and key is not to be dropped...
            if (vals
                and (real_key not in processed_keys)
                and not dropexp(real_key)
                and not orig_dropexp(real_key)):
                ## iterate vals
                for val in vals:
                    ## format literal '$'
                    if key.startswith('__mv'):
                        val = val.replace('$$', '$')
                    ## escape quotes
                    if isinstance(val, basestring):
                        val = val.replace('"', r'\"')
                    ## check map
                    if mapexp(real_key):
                        _raw += ', %s="%s"' % ('orig_' + real_key.lstrip('_'), val)
                    else:
                        _raw += ', %s="%s"' % (real_key, val)
                processed_keys.append(real_key)

        return _raw

    def addevent(self, raw, sourcetype, cam_header=True):
        """ The purpose of this method is to add a properly constructed event
        to the events list in the ModularAction instance.  This ensures events
        are created with the appropriate index-time header.

        The index-time header is responsible for setting sourcetype,
        orig_action_name, orig_sid, and orig_rid.  The index-time header will
        not be present in the _raw of generated events.

        Remember to call update() prior to addevent() to ensure that the events
        reflect the appropriate orig_sid and orig_rid identifiers.

        @param raw:        The text of the event you want to generate.
        @param sourcetype: The sourcetype of the event you want to generate.
        @param cam_header: Optionally exclude the inclusion of the index-time header.
                           Defaults to True (include header).
        """
        if cam_header:
            if self.orig_sid:
                action_idline = ModularAction.DEFAULT_IDLINE % (
                                    self.action_name,
                                    self.orig_sid,
                                    self.orig_rid,
                                    sourcetype)
            else:
                action_idline = ModularAction.DEFAULT_IDLINE % (
                                    self.action_name,
                                    self.sid,
                                    self.rid,
                                    sourcetype)
            self.events.append(action_idline + raw)
        else:
            self.events.append(raw)

    def writeevents(self, index='summary', host=None, source=None, fext='common_action_model'):
        """ The purpose of this method is to create arbitrary splunk events
        from the list of events in the ModularAction instance.

        Please use addevent() for populating the list of events in
        the ModularAction instance.

        @param index:  The index to write the events to.
                       Defaults to "summary".
        @param host:   The value of host the events should take on.
                       Defaults to None (auto).
        @param source: The value of source the events should take on.
                       Defaults to None (auto).
        @param fext:   The extension of the file to write out.
                       Files are written to $SPLUNK_HOME/var/spool/splunk.
                       File extensions can only contain word characters,
                       dash, and have a 200 char max.
                       "stash_" is automatically prepended to all extensions.
                       Defaults to "common_action_model" ("stash_common_action_model").
                       Only override if you've set up a corresponding props.conf
                       stanza to handle the extension.

        @return bool:  Returns True if all events were successfully written
                       Returns False if any errors were encountered
        """
        ## internal makeevents method for normalizing strings
        ## that will be used in the various headers we write out
        def get_string(input, default):
            try:
                return input.replace('"', '_')
            except AttributeError:
                return default

        if self.events:
            ## sanitize file extension
            if not fext or not re.match('^[\w-]+$', fext):
                self.logger.warn('Requested file extension was ignored due to invalid characters')
                fext = 'common_action_model'
            elif len(fext)>200:
                self.logger.warn('Requested file extension was ignored due to length')
                fext = 'common_action_model'
            ## header
            header_line = ModularAction.DEFAULT_HEADER % (
                            get_string(index, ModularAction.DEFAULT_INDEX),
                            get_string(host, ''),
                            get_string(source, ''))
            ## process event chunks
            for chunk in (self.events[x:x+ModularAction.DEFAULT_CHUNK]
                          for x in range(0, len(self.events), ModularAction.DEFAULT_CHUNK)):
                ## initialize output string
                default_breaker = '\n' + ModularAction.DEFAULT_BREAKER
                fout            = header_line + default_breaker + (default_breaker).join(chunk)
                ## write output string
                try:
                    fn = '%s_%s.stash_%s' % (mktimegm(time.gmtime()), random.randint(0, 100000), fext)
                    fp = make_splunkhome_path(['var', 'spool', 'splunk', fn])
                    ## obtain fh
                    with open(fp, 'w') as fh:
                        fh.write(fout)
                except:
                    signature = 'Error obtaining file handle during makeevents'
                    self.message(signature, level=logging.ERROR, file_path=fp)
                    self.logger.exception(signature + ' file_path=%s' % fp)
                    return False
            self.message('Successfully created splunk events', event_count=len(self.events))
            return True
        return False

    def dowork(self):
        """ This method serves as an illustration stub.
        Serves as a container for operations which satisfy the nature of the action.
        For instance, the third party API call.

        For cleanliness it is recommended that you subclass ModularAction
        and implement your own dowork() method.
        """
        return

    @staticmethod
    def setup_logger(name, level=logging.INFO, maxBytes=25000000, backupCount=5, format=SHORT_FORMAT):
        """ Set up a logging instance.

        @param name:        The log file name.
                            We recommend "$action_name$_modalert".
        @param level:       The logging level.
        @param maxBytes:    The maximum log file size before rollover.
        @param backupCount: The number of log files to retain.

        @return logger:     Returns an instance of logger
        """
        logfile = make_splunkhome_path(['var', 'log', 'splunk', name + '.log'])
        logger = logging.getLogger(name)
        logger.setLevel(level)
        logger.propagate = False  # Prevent the log messages from being duplicated in the python.log file

        # Prevent re-adding handlers to the logger object, which can cause duplicate log lines.
        handler_exists = any([True for h in logger.handlers if h.baseFilename == logfile])
        if not handler_exists:
            file_handler = logging.handlers.RotatingFileHandler(logfile, maxBytes=maxBytes, backupCount=backupCount)
            formatter = logging.Formatter(format)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

        return logger



"""
Converters for Splunk configuration.
"""

from __future__ import absolute_import

from past.builtins import basestring
import sys
from builtins import object
import base64
import json

basestring = str if sys.version_info[0] == 3 else basestring

__all__ = [
    'Converter',
    'Normaliser',
    'ChainOf',
    'UserDefined',
    'Unifier',
    'Boolean',
    'Lower',
    'Upper',
    'Mapping',
    'Base64',
    'JSON',
]


class Converter(object):
    """
    Converting data: encode for in-coming request
        and decode for out-coming response.
    """
    def encode(self, value, request):
        """
        Encode data from client for request.

        :param value: value to encode for request
        :param request: whole request data
        :return:
        """
        raise NotImplementedError()

    def decode(self, value, response):
        """
        Decode data from storage for response.

        :param value: value to decode for response
        :param response: whole response data
        :return:
        """
        raise NotImplementedError()


class Normaliser(Converter):
    """
    Normalizing data: same converting logic for encode & decode.
    """

    def normalize(self, value, data):
        """
        Normalize a given value.

        :param value: value to normalize
        :param data: whole payload
        :returns: normalized value.
        """
        raise NotImplementedError()

    def encode(self, value, request):
        return self.normalize(value, request)

    def decode(self, value, response):
        return self.normalize(value, response)


class ChainOf(Converter):
    """
    A composite of converters that will covert data with specified
    converters on by one, and returns result from the last converter.
    """

    def __init__(self, *converters):
        """

        :param converters: a list of converters
        """
        super(ChainOf, self).__init__()
        self._converters = converters

    def encode(self, value, request):
        for converter in self._converters:
            value = converter.encode(value, request)
        return value

    def decode(self, value, response):
        import copy
        converters = copy.copy(self._converters)
        converters.reverse()
        for converter in converters:
            value = converter.decode(value, response)
        return value


class UserDefined(Converter):
    """
    User-defined normaliser.

    The user-defined normaliser function should be in form:
    ``def fun(value, *args, **kwargs): ...``

    Usage::
    >>> def my_encoder(value, request, args):
    >>>     if request == args:
    >>>         return value
    >>>     else:
    >>>         return value
    >>> my_converter = UserDefined(my_encoder, 'test_val')
    >>> my_converter.encode('value', {'key': 'value'}, 'value1')

    """

    def __init__(self, encoder, decoder=None, *args, **kwargs):
        """

        :param encoder: user-defined function for encoding
        :param decoder: user-defined function for decoding.
            If None, it is the same to encoder.
        :param args:
        :param kwargs:
        """
        super(UserDefined, self).__init__()
        self._encoder = encoder
        self._decoder = decoder or self._encoder
        self._args = args
        self._kwargs = kwargs

    def encode(self, value, request):
        return self._encoder(
            value,
            request,
            *self._args,
            **self._kwargs
        )

    def decode(self, value, response):
        return self._decoder(
            value,
            response,
            *self._args,
            **self._kwargs
        )


class Lower(Normaliser):
    """
    Normalize a string to all lower cases.
    """

    def normalize(self, value, data):
        return value.strip().lower()


class Upper(Normaliser):
    """
    Normalize a string to all upper cases.
    """
    def normalize(self, value, data):
        return value.strip().upper()


class Unifier(Normaliser):
    """
    Many-to-one map for normalizing request & response.
    """

    def __init__(
            self,
            value_map,
            default=None,
            case_sensitive=False,
    ):
        """

        :param value_map:
            {"<unified value>": "<original value list>"}
        :param default: default value for input not in specific list
        :param case_sensitive: if it is False,
            it will return lower case
        """
        super(Unifier, self).__init__()
        self._case_sensitive = case_sensitive
        self._default = default
        self._value_map = {}
        for val_new, val_old_list in list(value_map.items()):
            for val_old in val_old_list:
                val_old = val_old if case_sensitive else val_old.lower()
                assert val_old not in self._value_map, \
                    'Normaliser "Unifier" only supports Many-to-one mapping: %s' % val_old
                self._value_map[val_old] = val_new

    def normalize(self, value, data):
        need_lower = not self._case_sensitive and \
                     isinstance(value, basestring)
        val_old = value.lower() if need_lower else value
        val_default = self._default or value
        return self._value_map.get(val_old, val_default)


class Boolean(Unifier):
    """
    Normalize a boolean field.

    Normalize given value to boolean: 0 or 1
        (for False and True respectively).
    If the given value is not-a-string or unrecognizable,
    it returns default value.
    """

    VALUES_TRUE = {'true', 't', '1', 'yes', 'y'}
    VALUES_FALSE = {'false', 'f', '0', 'no', 'n'}

    def __init__(self, default=True):
        """

        :param default: default for unrecognizable input of boolean.
        """
        super(Boolean, self).__init__(
            value_map={
                '1': Boolean.VALUES_TRUE,
                '0': Boolean.VALUES_FALSE,
            },
            default='1' if default else '0',
            case_sensitive=False,
        )


class Mapping(Converter):
    """
    One-to-one map between interface value and storage value.
    If value is not in specific mapping,
    it will return the original value.
    """

    def __init__(self, value_map, case_sensitive=False):
        """

        :param value_map: {"<interface value>": "<storage value>"}
        :param case_sensitive: if it is False,
            it will return lower case
        """
        super(Mapping, self).__init__()
        self._case_sensitive = case_sensitive
        self._map_interface, self._map_storage = {}, {}
        for interface, storage in list(value_map.items()):
            self._check_and_set(interface, storage)

    def _check_and_set(self, interface, storage):
        if not self._case_sensitive:
            interface = interface.lower()
            storage = storage.lower()
        assert interface not in self._map_interface, \
            'Converter "Mapping" only supports one-to-one mapping: "%s"' % interface
        assert storage not in self._map_storage, \
            'Converter "Mapping" only supports one-to-one mapping: "%s"' % storage
        self._map_interface[interface] = storage
        self._map_storage[storage] = interface

    def encode(self, value, request):
        if self._case_sensitive:
            interface = value
        else:
            interface = value.lower()
        return self._map_interface.get(interface, value)

    def decode(self, value, response):
        if self._case_sensitive:
            storage = value
        else:
            storage = value.lower()
        return self._map_storage.get(storage, value)


class Base64(Converter):
    """
    Covert input data to base64 string.
    """

    def encode(self, value, request):
        return base64.b64encode(value)

    def decode(self, value, response):
        return base64.b64decode(value)


class JSON(Converter):
    """
    Converter between object and JSON string.
    """

    def encode(self, value, request):
        return json.dumps(value)

    def decode(self, value, response):
        return json.loads(value)



SHORT_NAME = "short_name"
TITLE = "title"
PARAMETERS = "parameters"
NAME = "name"
FORMAT_TYPE = "format_type"
REQUIRED = "required"
DEFAULT_VALUE = "default_value"
LABEL = "label"
HELP_STRING = "help_string"
HELP_LINK = "help_link"
FORM_CLASS = "form_class"
FORM_STYLE = "form_style"
POSSIBLE_VALUES = "possible_values"
MODULAR_ALERTS = "modular_alerts"
PLACEHOLDER = "placeholder"
LINK_TEXT = "link_text"
LINK_URL = "link_url"
LINK_TIP = "link_tip"
LINK_TYPE = "link_type"
ICON_PATH = "icon_path"
ALERT_PROPS = "alert_props"
RESULTS = "results"
RESULT = "result"
RESULTS_FILE = "results_file"
STDIN_FIELDS = "stdin_fields"
LARGE_ICON = "largeIcon"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['splunk_add_on_ucc_framework',
 'splunk_add_on_ucc_framework.alert_utils',
 'splunk_add_on_ucc_framework.alert_utils.alert_utils_common',
 'splunk_add_on_ucc_framework.alert_utils.alert_utils_common.metric_collector',
 'splunk_add_on_ucc_framework.arf_dir_templates.modular_alert_package.${product_id}.lib',
 'splunk_add_on_ucc_framework.arf_dir_templates.modular_alert_package.${product_id}.lib.splunk_aoblib',
 'splunk_add_on_ucc_framework.modular_alert_builder',
 'splunk_add_on_ucc_framework.modular_alert_builder.build_core',
 'splunk_add_on_ucc_framework.splunktaucclib',
 'splunk_add_on_ucc_framework.splunktaucclib.global_config',
 'splunk_add_on_ucc_framework.splunktaucclib.rest_handler',
 'splunk_add_on_ucc_framework.splunktaucclib.rest_handler.endpoint',
 'splunk_add_on_ucc_framework.uccrestbuilder',
 'splunk_add_on_ucc_framework.uccrestbuilder.endpoint']

package_data = \
{'': ['*'],
 'splunk_add_on_ucc_framework': ['arf_dir_templates/*',
                                 'arf_dir_templates/modular_alert_package/${product_id}/appserver/static/*',
                                 'templates/*'],
 'splunk_add_on_ucc_framework.modular_alert_builder.build_core': ['arf_template/*',
                                                                  'arf_template/default_html_theme/*']}

install_requires = \
['future>=0,<1',
 'jinja2>=2,<3',
 'lxml>=4.3,<5.0',
 'mako>=1,<2',
 'munch>=2,<3',
 'solnlib>=2.1,<3.0',
 'splunk-sdk>=1.6,<2.0']

entry_points = \
{'console_scripts': ['build-ucc = build:build_ucc',
                     'install-libs = '
                     'splunk_add_on_ucc_framework:install_requirements',
                     'ucc-gen = splunk_add_on_ucc_framework:main']}

setup_kwargs = {
    'name': 'splunk-add-on-ucc-framework',
    'version': '4.0.4',
    'description': 'Splunk Add-on SDK formerly UCC is a build and code generation framework',
    'long_description': '',
    'author': 'rfaircloth-splunk',
    'author_email': 'rfaircloth@splunk.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/splunk/splunk-add-on-sdk-python/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
